from operator import add, eq, mul, pow, sub, truediv

from ....datalog import Conjunction, Fact, Implication, Negation, Union
from ....datalog.aggregation import AggregationApplication
from ....expressions import Constant, Symbol
from ..standard_syntax import ExternalSymbol
from ..natural_syntax import parser
from ....probabilistic.expressions import ProbabilisticPredicate


def test_facts():
    res = parser('A(3)')
    assert res == Union((Fact(Symbol('A')(Constant(3.))),))

    res = parser('A("x")')
    assert res == Union((Fact(Symbol('A')(Constant('x'))),))

    res = parser("A('x', 3)")
    assert res == Union((Fact(Symbol('A')(Constant('x'), Constant(3.))),))

    res = parser(
        'A("x", 3)\n'
        'ans():-A(x, y)'
    )
    assert res == Union((
        Fact(Symbol('A')(Constant('x'), Constant(3.))),
        Implication(
            Symbol('ans')(),
            Conjunction((
                Symbol('A')(Symbol('x'), Symbol('y')),
            ))
        )
    ))

    res = parser('"john" is cat')
    assert res == Union((Fact(Symbol('cat')(Constant("john"))),))

    res = parser('"john" is `http://owl#fact`')
    assert res == Union((Fact(Symbol('http://owl#fact')(Constant("john"))),))

    res = parser('"john" is "perceval"\'s mascot')
    assert res == Union((
        Fact(Symbol('mascot')(Constant("john"), Constant("perceval"))),
    ))

    res = parser('"john" has 4 legs')
    assert res == Union((
        Fact(Symbol('legs')(Constant("john"), Constant(4.))),
    ))

    res = parser('"john" has 4 legs')
    assert res == Union((
        Fact(Symbol('legs')(Constant("john"), Constant(4.))),
    ))

    res = parser('"john" is below the "table"')
    assert res == Union((
        Fact(Symbol('below')(Constant("john"), Constant("table"))),
    ))


def test_rules():
    A = Symbol('A')
    B = Symbol('B')
    C = Symbol('C')
    f = Symbol('f')
    x = Symbol('x')
    y = Symbol('y')
    z = Symbol('z')
    res = parser('A(x):-B(x, y), C(3, z)')
    assert res == Union((
        Implication(A(x), Conjunction((B(x, y), C(Constant(3), z)))),
    ))

    res = parser('A(x):-~B(x)')
    assert res == Union((
        Implication(A(x), Conjunction((Negation(B(x)),))),
    ))

    res = parser('A(x):-B(x, y), C(3, z), z == 4')
    assert res == Union((
        Implication(
            A(x),
            Conjunction((
                B(x, y), C(Constant(3), z), Constant(eq)(z, Constant(4.))
            ))
        ),
    ))

    res = parser('A(x):-B(x + 5 * 2, y), C(3, z), z == 4')
    assert res == Union((
        Implication(
            A(x),
            Conjunction((
                B(
                    Constant(add)(
                        x,
                        Constant(mul)(Constant(5.), Constant(2.))),
                    y
                ),
                C(Constant(3), z), Constant(eq)(z, Constant(4.))
            ))
        ),
    ))

    res = parser('A(x):-B(x / 2, y), C(3, z), z == 4')
    assert res == Union((
        Implication(
            A(x),
            Conjunction((
                B(
                    Constant(truediv)(x, Constant(2.)),
                    y
                ),
                C(Constant(3), z), Constant(eq)(z, Constant(4.))
            ))
        ),
    ))

    res = parser('A(x):-B(f(x), y), C(3, z), z == 4')
    assert res == Union((
        Implication(
            A(x),
            Conjunction((
                B(f(x), y),
                C(Constant(3), z), Constant(eq)(z, Constant(4.))
            ))
        ),
    ))

    res = parser('A(x):-B(x + (-5), "a")')
    assert res == Union((
        Implication(
            A(x),
            Conjunction((
                B(
                    Constant(add)(x, Constant(-5.)),
                    Constant("a")
                ),
            ))
        ),
    ))

    res = parser('A(x):-B(x - 5 * 2, @y ** -2)')
    assert res == Union((
        Implication(
            A(x),
            Conjunction((
                B(
                    Constant(sub)(
                        x,
                        Constant(mul)(Constant(5.), Constant(2.))
                    ),
                    Constant(pow)(ExternalSymbol('y'), Constant(-2.))
                ),
            ))
        ),
    ))


def test_nl_rules():
    cat = Symbol('cat')
    bird = Symbol('bird')
    feline = Symbol('feline')
    legs = Symbol('legs')
    small = Symbol('small')
    goodluck_cat = Symbol('goodluck_cat')
    black = Symbol('black')

    x = Symbol('x')
    y = Symbol('y')

    res = parser('x is cat if x is feline, x has 4 legs, x is small')
    assert res == Union((
        Implication(
            cat(x), Conjunction((feline(x), legs(x, Constant(4.)), small(x)))
        ),
    ))

    res = parser('x is goodluck_cat if x is cat, not x is black')
    assert res == Union((
        Implication(
            goodluck_cat(x), Conjunction((cat(x), Negation(black(x))))
        ),
    ))

    res = parser('''
        x has y legs if x is a cat & y == 4.0
        or x has y legs if x is a bird and y == 2
    ''')
    assert res == Union((
        Implication(
            legs(x, y), Conjunction((cat(x), Constant(eq)(y, Constant(4.))))
        ),
        Implication(
            legs(x, y), Conjunction((bird(x), Constant(eq)(y, Constant(2))))
        ),
    ))


def test_aggregation():
    A = Symbol('A')
    B = Symbol('B')
    f = Symbol('f')
    x = Symbol('x')
    y = Symbol('y')
    res = parser('A(x, f(y)):-B(x, y)')
    assert res == Union((
        Implication(
            A(x, AggregationApplication(f, (y,))),
            Conjunction((B(x, y),))
        ),
    ))


def test_aggregation_nsd():
    A = Symbol('A')
    B = Symbol('B')
    f = Symbol('f')
    x = Symbol('x')
    y = Symbol('y')
    res = parser('x has f(y) A if B(x, y)')
    assert res == Union((
        Implication(
            A(x, AggregationApplication(f, (y,))),
            Conjunction((B(x, y),))
        ),
    ))


def test_uri_nsd():
    from rdflib import RDFS

    label = Symbol(name=str(RDFS.label))
    regional_part = Symbol(name='http://sig.biostr.washington.edu/fma3.0#regional_part_of')
    x = Symbol('x')

    res = parser(f'x is `{str(label.name)}` if x is `{str(regional_part.name)}`')
    expected_result = Union((
        Implication(
            label(x),
            Conjunction((regional_part(x),))
        ),
    ))

    assert res == expected_result


def test_probabilistic_fact():
    A = Symbol('A')
    p = Symbol('p')
    res = parser('p::A(3)')
    assert res == Union((
        Implication(
            ProbabilisticPredicate(p, A(Constant(3.))),
            Constant(True)
        ),
    ))

    res = parser('0.8::A("a b", 3)')
    assert res == Union((
        Implication(
            ProbabilisticPredicate(
                Constant(0.8),
                A(Constant("a b"), Constant(3.))
            ),
            Constant(True)
        ),
    ))


def test_probabilistic_fact_nsd():
    cat = Symbol('cat')
    p = Symbol('p')
    res = parser('with probability p "john" is cat')
    assert res == Union((
        Implication(
            ProbabilisticPredicate(p, cat(Constant("john"))),
            Constant(True)
        ),
    ))

    res = parser("with probability p 'john' is 'george''s cat")
    assert res == Union((
        Implication(
            ProbabilisticPredicate(p, cat(Constant("john"), Constant("george"))),
            Constant(True)
        ),
    ))



import numpy as np


def before(i1, i2) -> bool:
    return i1[0] < i1[1] < i2[0] < i2[1]


def overlaps(i1, i2) -> bool:
    return i1[0] < i2[0] < i1[1] < i2[1]


def during(i1, i2) -> bool:
    return i2[0] < i1[0] < i1[1] < i2[1]


def meets(i1, i2) -> bool:
    return i1[0] < i1[1] == i2[0] < i2[1]


def starts(i1, i2) -> bool:
    return i1[0] == i2[0] < i1[1] < i2[1]


def finishes(i1, i2) -> bool:
    return i2[0] < i1[0] < i1[1] == i2[1]


def equals(i1, i2) -> bool:
    return i1[0] == i2[0] < i1[1] == i2[1]


def converse(operation):
    return lambda x, y: operation(y, x)


def negate(operation):
    return lambda x, y: not operation(x, y)


def get_intervals_relations(intervals, other_region_intervals):
    obtained_relation_per_axis = [''] * len(intervals)
    relations = [before, overlaps, during, meets, starts, finishes, equals]
    for f in relations:
        for i, _ in enumerate(obtained_relation_per_axis):
            if f(intervals[i], other_region_intervals[i]):
                obtained_relation_per_axis[i] = str(f.__name__[0])
            elif f(other_region_intervals[i], intervals[i]):
                obtained_relation_per_axis[i] = str(f.__name__[0] + 'i')
        if np.all(np.array(obtained_relation_per_axis) != ''):
            break
    return tuple(obtained_relation_per_axis)


def v_before(i1, i2):
    if i1[0] < i1[1] < i2[0] < i2[1]:
        return (1, 0, 0)
    elif i2[0] < i2[1] < i1[0] < i1[1]:
        return (0, 0, 1)
    return None


def v_overlaps(i1, i2):
    if i1[0] < i2[0] < i1[1] < i2[1]:
        return (1, 1, 0)
    elif i2[0] < i1[0] < i2[1] < i1[1]:
        return (0, 1, 1)
    return None


def v_during(i1, i2):
    if i2[0] < i1[0] < i1[1] < i2[1]:
        return (0, 1, 0)
    elif i1[0] < i2[0] < i2[1] < i1[1]:
        return (1, 1, 1)
    return None


def v_meets(i1, i2):
    if i1[0] < i1[1] == i2[0] < i2[1]:
        return (1, 0, 0)
    elif i2[0] < i2[1] == i1[0] < i1[1]:
        return (0, 0, 1)
    return None


def v_starts(i1, i2):
    if i1[0] == i2[0] < i1[1] < i2[1]:
        return (0, 1, 0)
    elif i2[0] == i1[0] < i2[1] < i1[1]:
        return (0, 1, 1)
    return None


def v_finishes(i1, i2):
    if i2[0] < i1[0] < i1[1] == i2[1]:
        return (0, 1, 0)
    elif i1[0] < i2[0] < i2[1] == i1[1]:
        return (1, 1, 0)


def v_equals(i1, i2):
    if i1[0] == i2[0] < i1[1] == i2[1]:
        return (0, 1, 0)
    return None



"""
Expressions for the intermediate representation of a
Datalog program.
"""

from typing import Any

from ..expression_walker import add_match
from ..expressions import Constant, ExpressionBlock
from ..logic import (Conjunction, Implication, Negation, UnaryLogicOperator,
                     Union)
from ..logic import expression_processing as ep

__all__ = [
    "Fact", "TranslateToLogic", "Implication", "Conjunction", "Union",
    "Negation", "NULL", "UNDEFINED", "UnaryLogicOperator"
]


class Fact(Implication):
    def __init__(self, consequent):
        super().__init__(consequent, Constant(True))

    @property
    def fact(self):
        return self.consequent

    def __repr__(self):
        return 'Fact{{{} \u2190 {}}}'.format(
            repr(self.consequent), True
        )


class Undefined(Constant):
    def __repr__(self):
        return 'UNDEFINED'


class NullConstant(Constant):
    def __repr__(self):
        return 'NULL'


UNDEFINED = Undefined(None)
NULL = NullConstant[Any](None)


class TranslateToLogic(ep.TranslateToLogic):
    @add_match(
        Implication(..., Constant(True)),
        lambda x: not isinstance(x, Fact)
    )
    def translate_true_implication(self, implication):
        return self.walk(Fact(implication.consequent))

    @add_match(ExpressionBlock)
    def build_conjunction_from_expression_block(self, expression_block):
        formulas = tuple()
        for expression in expression_block.expressions:
            new_exp = self.walk(expression)
            formulas += (new_exp,)
        return self.walk(Union(formulas))



from ....datalog import Fact
from ....expressions import Constant, Symbol
from ....logic import Conjunction, Implication, Union
from .. import testing
from ..cplogic_to_gm import (
    AndPlateNode,
    BernoulliPlateNode,
    NaryChoicePlateNode,
    NaryChoiceResultPlateNode,
)
from ..program import CPLogicProgram

P = Symbol("P")
Q = Symbol("Q")
Z = Symbol("Z")
x = Symbol("x")
a = Constant("a")
b = Constant("b")


def test_empty_program():
    code = Union(tuple())
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    gm = testing.build_gm(cpl_program)
    assert gm.edges == tuple()
    assert gm.nodes == tuple()


def test_simple_deterministic_program():
    code = Union((Fact(Q(a)), Fact(Q(b)), Implication(P(x), Q(x))))
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    gm = testing.build_gm(cpl_program)
    assert len(gm.nodes) == 2
    Q_node = gm.get_node(Q)
    P_node = gm.get_node(P)
    assert isinstance(Q_node, BernoulliPlateNode)
    assert isinstance(P_node, AndPlateNode)
    assert len(Q_node.relation.value) == 2
    assert Q_node.relation.value.arity == 2
    assert len(P_node.relation.value) == 2
    assert P_node.relation.value.arity == 1


def test_program_with_probfacts():
    code = Union((Implication(Q(x), P(x)),))
    probfacts_sets = {P: {(1.0, "a"), (0.5, "b"), (0.3, "c")}}
    cpl_program = CPLogicProgram()
    cpl_program.walk(code)
    for pred_symb, pfact_set in probfacts_sets.items():
        cpl_program.add_probabilistic_facts_from_tuples(pred_symb, pfact_set)
    gm = testing.build_gm(cpl_program)
    assert len(gm.edges) == 1
    Q_node = gm.get_node(Q)
    P_node = gm.get_node(P)
    assert gm.get_parent_node_symbols(Q) == {P}
    assert isinstance(Q_node, AndPlateNode)
    assert isinstance(P_node, BernoulliPlateNode)
    assert (
        testing.get_named_relation_tuples(P_node.relation) == probfacts_sets[P]
    )


def test_program_with_probchoice():
    cpl_program = CPLogicProgram()
    cpl_program.add_probabilistic_choice_from_tuples(
        P, {(0.6, "a"), (0.4, "b")}
    )
    gm = testing.build_gm(cpl_program)
    P_node = gm.get_node(P)
    assert isinstance(P_node, NaryChoiceResultPlateNode)
    assert len(gm.get_parent_node_symbols(P)) == 1
    choice_node_symb = next(iter(gm.get_parent_node_symbols(P)))
    choice_node = gm.get_node(choice_node_symb)
    assert isinstance(choice_node, NaryChoicePlateNode)


def test_program_with_probchoice_and_intensional_rule():
    cpl_program = CPLogicProgram()
    cpl_program.walk(Union((Implication(Q(x), Conjunction((P(x), Z(x)))),)))
    cpl_program.add_probabilistic_facts_from_tuples(
        Z, {(0.6, "a"), (1.0, "b")}
    )
    cpl_program.add_probabilistic_choice_from_tuples(
        P, {(0.6, "a"), (0.4, "b")}
    )
    gm = testing.build_gm(cpl_program)
    assert gm.get_parent_node_symbols(Q) == {P, Z}
    assert len(gm.get_parent_node_symbols(P)) == 1
    choice_node_symb = next(iter(gm.get_parent_node_symbols(P)))
    Z_node = gm.get_node(Z)
    P_node = gm.get_node(P)
    choice_node = gm.get_node(choice_node_symb)
    assert isinstance(Z_node, BernoulliPlateNode)
    assert isinstance(P_node, NaryChoiceResultPlateNode)
    assert isinstance(choice_node, NaryChoicePlateNode)



import numpy as np
import pytest

from ...exceptions import NeuroLangNotImplementedError
from ..distributions import Distribution, TableDistribution
from ..exceptions import UncomparableDistributionsError

dog_cat_table = {"cat": 0.2, "dog": 0.8}

beach_jellyfish_table = {
    "chironex_fleckeri": 0.01,
    "portuguese_man_o_war": 0.05,
    "pelagia_noctiluca": 0.1,
    "lion_mane": 0.14,
    "chrysaora_hysoscella": 0.2,
    "moon_jelly": 0.2,
    "aequorea_victoria": 0.3,
}

jellyfish_dangerosity = {
    "chironex_fleckeri": 1.0,
    "portuguese_man_o_war": 0.8,
    "pelagia_noctiluca": 0.4,
    "lion_mane": 0.4,
    "chrysaora_hysoscella": 0.1,
    "moon_jelly": 0.05,
    "aequorea_victoria": 0.0,
}


def test_abstract_distribution_class():
    with pytest.raises(NeuroLangNotImplementedError):
        _ = Distribution().probability(42)
    with pytest.raises(NeuroLangNotImplementedError):
        _ = Distribution().support
    with pytest.raises(NeuroLangNotImplementedError):
        _ = Distribution().expectation(lambda x: x)


def test_table_distribution():
    distrib = TableDistribution(dog_cat_table)
    assert distrib.probability("cat") == 0.2
    assert distrib.probability("dog") == 0.8
    assert distrib.support == frozenset({"cat", "dog"})


def test_expectation_table_distribution():
    beach_jellyfish_dist = TableDistribution(beach_jellyfish_table)
    beach_dangerosity = beach_jellyfish_dist.expectation(
        lambda v: jellyfish_dangerosity[v]
    )
    expected_value = sum(
        jellyfish_dangerosity[k] * beach_jellyfish_table[k]
        for k in jellyfish_dangerosity.keys()
    )
    assert beach_dangerosity == expected_value


def test_conditional_table_distribution():
    observed_jellyfishes = {"chironex_fleckeri", "moon_jelly"}
    condition = lambda value: value in observed_jellyfishes
    beach_jellyfish_dist = TableDistribution(beach_jellyfish_table)
    beach_jellyfish_dist = beach_jellyfish_dist.conditioned_on(condition)
    beach_dangerosity = beach_jellyfish_dist.expectation(
        lambda v: jellyfish_dangerosity[v]
    )
    assert np.isclose(beach_dangerosity, 0.09523809523809525)


def test_compare_table_distribs():
    td1 = TableDistribution(beach_jellyfish_table)
    assert td1 == td1
    td2_table = beach_jellyfish_table.copy()
    td2_table["pelagia_noctiluca"] = 0.14
    td2_table["lion_mane"] = 0.1
    td2 = TableDistribution(td2_table)
    assert td1 != td2


def test_compare_table_distrib_with_non_table_distrib():
    td = TableDistribution(beach_jellyfish_table)
    with pytest.raises(UncomparableDistributionsError):
        if td == Distribution():
            pass



import typing
import re

from .CD_relations import cardinal_relation, inverse_directions
from .regions import Region, region_union
from .expression_walker import PatternWalker
from .expressions import Constant


REFINE_OVERLAPPING = True


class RegionSolver(PatternWalker[Region]):
    type_name = 'Region'

    def __new__(cls, *args, **kwargs):
        cardinal_operations = {
            'inferior_of': 'I', 'superior_of': 'S',
            'posterior_of': 'P', 'anterior_of': 'A',
            'left_of': 'L', 'right_of': 'R',
            'overlapping': 'O'
        }

        refine_overlapping = kwargs.get(
            'refine_overlapping',
            REFINE_OVERLAPPING
        )

        max_tree_depth_level = kwargs.get(
            'max_tree_depth_level',
            None
        )

        def build_function(relation, refine_overlapping=False):
            def fun(self, x: Region, y: Region) -> bool:
                return bool(cardinal_relation(
                    x, y, relation,
                    refine_overlapping=refine_overlapping,
                    stop_at=max_tree_depth_level
                ))
            return fun

        def anatomical_direction_function(relation, refine_overlapping=False):

            def func(self, x: Region, y: Region) -> bool:

                return bool(
                    cardinal_relation(
                        x, y, relation,
                        refine_overlapping=refine_overlapping,
                        stop_at=max_tree_depth_level
                    ) and not (
                        cardinal_relation(
                            x, y, inverse_directions[relation],
                            refine_overlapping=refine_overlapping,
                            stop_at=max_tree_depth_level
                        ) or
                        cardinal_relation(
                            x, y, cardinal_operations['overlapping'],
                            refine_overlapping=refine_overlapping,
                            stop_at=max_tree_depth_level
                        )
                    )
                )

            return func

        for key, value in cardinal_operations.items():
            setattr(
                cls, f'function_{key}',
                build_function(value, refine_overlapping=refine_overlapping)
            )

        anatomical_correct_operations = {
            k: cardinal_operations[k] for k in (
                'inferior_of', 'superior_of',
                'posterior_of', 'anterior_of'
                )
        }
        for key, value in anatomical_correct_operations.items():
            setattr(
                cls, f'function_anatomical_{key}',
                anatomical_direction_function(
                    value, refine_overlapping=refine_overlapping
                )
            )

        return PatternWalker.__new__(cls)

    def function_regexp(
        self, regexp: typing.Text
    ) -> typing.AbstractSet[Region]:
        regions = []
        for k in self.symbol_table.symbols_by_type(Region):
            if re.search(regexp, k.name):
                regions.append(k)

        return frozenset(regions)

    def function_region_union(
        self, region_set: typing.AbstractSet[Region]
    ) -> Region:

        new_region_set = []
        for region in region_set:
            region = self.walk(region)
            if not isinstance(region, Constant):
                raise ValueError(
                    "Region union can only be evaluated on resolved regions"
                )

            new_region_set.append(region.value)

        return region_union(new_region_set)



import operator

import pytest

from ...exceptions import UnsupportedProgramError
from ...expression_walker import ExpressionWalker
from ...expressions import Constant, Symbol
from ...logic import (
    FALSE,
    Conjunction,
    Disjunction,
    ExistentialPredicate,
    Implication,
    Union,
)
from ..basic_representation import DatalogProgram
from ..expression_processing import extract_logic_predicates, flatten_query

EQ = Constant(operator.eq)

P = Symbol("P")
R = Symbol("R")
Z = Symbol("Z")
Q = Symbol("Q")
x = Symbol("x")
y = Symbol("y")
z = Symbol("z")

a = Constant("a")
b = Constant("b")


class TestDatalogProgram(DatalogProgram, ExpressionWalker):
    pass


def test_flatten_extensional_predicate_query():
    result = flatten_query(P(x), TestDatalogProgram())
    assert result == Conjunction((P(x),))


def test_flatten_with_renaming():
    code = Union(
        (
            Implication(Z(x), P(x)),
            Implication(R(y), P(y)),
            Implication(Q(y), Conjunction((Z(y), R(x)))),
        )
    )
    program = TestDatalogProgram()
    program.walk(code)
    result = flatten_query(Q(x), program)
    assert len(result.formulas) == 2
    assert P(x) in result.formulas
    assert any(
        pred.functor == P and pred.args[0] != x for pred in result.formulas
    )


def test_flatten_query_multiple_preds():
    query = Conjunction((P(x), Z(y)))
    result = flatten_query(query, TestDatalogProgram())
    assert len(result.formulas) == 2
    assert P(x) in result.formulas
    assert Z(y) in result.formulas


def test_existential():
    code = Union(
        (
            Implication(P(x), Conjunction((Z(x, y), Z(y, x)))),
            Implication(Q(x), Conjunction((P(x), P(y), R(z)))),
        )
    )
    program = TestDatalogProgram()
    program.walk(code)
    query = Q(z)
    result = flatten_query(query, program)
    assert any(
        formula.functor == R and formula.args[0] != z
        for formula in result.formulas
    )
    assert any(
        formula.functor == Z
        and formula.args[0] == z
        and "fresh" in formula.args[1].name
        for formula in result.formulas
    )
    match = next(
        formula
        for formula in result.formulas
        if formula.functor == Z
        and "fresh" in formula.args[0].name
        and "fresh" in formula.args[1].name
    )
    expected = match.functor(*reversed(match.args))
    assert expected in result.formulas


def test_flatten_with_top_level_disjunction():
    code = Union(
        (
            Implication(R(x), Z(x)),
            Implication(R(y), P(y)),
            Implication(Q(x), R(x)),
        )
    )
    program = TestDatalogProgram()
    program.walk(code)
    result = flatten_query(Q(x), program)
    assert isinstance(result, Disjunction)
    assert len(result.formulas) == 2
    assert P(x) in result.formulas
    assert Z(x) in result.formulas


def test_flatten_with_2nd_level_disjunction():
    code = Union(
        (
            Implication(R(x), Z(x)),
            Implication(R(y), P(y)),
            Implication(Q(x, z), Conjunction((R(x), Z(z)))),
        )
    )
    program = TestDatalogProgram()
    program.walk(code)
    result = flatten_query(Q(x, z), program)
    assert isinstance(result, Conjunction)
    assert len(result.formulas) == 2
    assert (
        isinstance(result.formulas[0], Disjunction)
        and set((P(x), Z(x))) == set(result.formulas[0].formulas)
    ) or (
        isinstance(result.formulas[1], Disjunction)
        and set((P(x), Z(x))) == set(result.formulas[1].formulas)
    )
    assert Z(z) == result.formulas[0] or Z(z) == result.formulas[1]


def test_flatten_recursive():
    code = Union(
        (
            Implication(R(x, y), Conjunction((Z(y), Z(x)))),
            Implication(R(x, y), Conjunction((Z(x), R(x, y)))),
        )
    )
    program = TestDatalogProgram()
    program.walk(code)
    with pytest.raises(UnsupportedProgramError):
        flatten_query(R(x, y), program)


def test_incorrect_program():
    code = Union(
        (
            Implication(
                R(x), ExistentialPredicate(y, Conjunction((Z(y), Z(x))))
            ),
        )
    )
    program = TestDatalogProgram()
    program.walk(code)
    assert flatten_query(R(x, y), program) == FALSE


def test_flatten_with_variable_equality():
    rule = Implication(R(x, y), Conjunction((Z(x), P(x, y), EQ(x, a))))
    program = TestDatalogProgram()
    program.walk(rule)
    flat = flatten_query(R(x, y), program)
    assert set(flat.formulas) == set(extract_logic_predicates(rule.antecedent))


def test_flatten_repeated_variable_in_rule():
    """
    Given a program with the rule `R(x, x) :- Q(x, x, z)` and a (single
    predicate) conjunctive query `R(x, y)`, we expect the resulting flattened
    query to be either `Q(y, y, _freshvar_), x = y` or `Q(x, x, _freshvar_), y
    = x`.

    """
    rule = Implication(R(x, x), Q(x, x, z))
    program = TestDatalogProgram()
    program.walk(rule)
    flat = flatten_query(R(x, y), program)
    assert len(flat.formulas) == 2
    assert (
        Constant(operator.eq)(x, y) in flat.formulas
        and any(
            (formula.functor == Q and formula.args[:2] == (y, y))
            for formula in flat.formulas
        )
    ) or (
        Constant(operator.eq)(y, x) in flat.formulas
        and any(
            (formula.functor == Q and formula.args[:2] == (x, x))
            for formula in flat.formulas
        )
    )


def test_flatten_not_unifiable_becomes_false():
    code = Union(
        (
            Implication(R(Constant(2), Constant(3)), Conjunction((Q(y, x),))),
            Implication(Z(x), Conjunction((R(x, x),))),
        )
    )
    program = TestDatalogProgram()
    program.walk(code)
    flat = flatten_query(Z(z), program)
    assert flat == FALSE


def test_flatten_query_double_vareq():
    code = Implication(P(x), Conjunction((R(x),)))
    program = TestDatalogProgram()
    program.walk(code)
    query = Conjunction((P(a), P(b)))
    flat = flatten_query(query, program)
    assert isinstance(flat, Conjunction)
    assert set(flat.formulas) == {R(a), R(b)}



from collections import namedtuple
from itertools import chain, tee
import logging
from operator import contains, eq
from typing import Iterable, Tuple

from ...exceptions import NeuroLangException
from ...expressions import Constant, FunctionApplication, Symbol
from ...logic.unification import (apply_substitution,
                                  apply_substitution_arguments,
                                  compose_substitutions)
from ...type_system import (NeuroLangTypeException, Unknown, get_args,
                            is_leq_informative, unify_types)
from ...utils import OrderedSet, log_performance
from ..expression_processing import (extract_logic_free_variables,
                                     extract_logic_predicates, is_linear_rule,
                                     dependency_matrix, program_has_loops)
from ..instance import MapInstance


LOG = logging.getLogger(__name__)


ChaseNode = namedtuple('ChaseNode', 'instance children')


class NeuroLangNonLinearProgramException(NeuroLangException):
    pass


class NeuroLangProgramHasLoopsException(NeuroLangException):
    pass


class ChaseGeneral():
    """Chase implementation using the naive resolution algorithm.

    """
    def __init__(self, datalog_program, rules=None):
        self.datalog_program = datalog_program
        self._set_rules(rules)

        self.builtins = datalog_program.builtins()
        self.idb_edb_symbols = set(
            chain(
                self.datalog_program.extensional_database(),
                self.datalog_program.intensional_database()
            )
        ) | set(
            rule.consequent.functor
            for rule in self.rules
        )

    def _set_rules(self, rules):
        self.rules = []
        if rules is None:
            for union in \
                    self.datalog_program.intensional_database().values():
                self.rules += union.formulas
        else:
            self.rules += rules.formulas

    def check_constraints(self, instance_update):
        pass

    def chase_step(self, instance, rule, restriction_instance=None):
        if restriction_instance is None:
            restriction_instance = MapInstance()

        rule_predicates = self.extract_rule_predicates(
            rule, instance, restriction_instance=restriction_instance
        )

        if all(len(predicate_list) == 0 for predicate_list in rule_predicates):
            return MapInstance()

        restricted_predicates, nonrestricted_predicates, builtin_predicates =\
            rule_predicates

        builtin_predicates, builtin_predicates_ = tee(builtin_predicates)
        args_to_project = self.get_args_to_project(rule, builtin_predicates_)

        rule_predicates_iterator = chain(
            restricted_predicates, nonrestricted_predicates
        )

        substitutions = self.obtain_substitutions(
            args_to_project, rule_predicates_iterator
        )

        substitutions = self.eliminate_already_computed(
            rule.consequent, instance, substitutions
        )

        substitutions = self.evaluate_builtins(
            builtin_predicates, substitutions
        )

        return self.compute_result_set(
            rule, substitutions, instance, restriction_instance
        )

    def get_args_to_project(self, rule, builtin_predicates_):
        args_to_project = self.extract_variable_arguments(rule.consequent)
        for predicate, _ in builtin_predicates_:
            args_to_project |= self.extract_variable_arguments(predicate)
        new_args_to_project = OrderedSet()
        for a in args_to_project:
            new_args_to_project.add(a)
        args_to_project = new_args_to_project
        return args_to_project

    @staticmethod
    def extract_variable_arguments(predicate):
        return extract_logic_free_variables(predicate)

    def evaluate_builtins(self, builtin_predicates, substitutions):
        new_substitutions = []
        predicates = [p for p, _ in builtin_predicates]
        if len(predicates) == 0:
            return substitutions
        for substitution in substitutions:
            updated_substitutions = self.evaluate_builtins_predicates(
                predicates, substitution
            )
            new_substitutions += updated_substitutions
        return new_substitutions

    def evaluate_builtins_predicates(
        self, predicates_to_evaluate, substitution
    ):
        substitutions = [substitution]
        predicates_to_evaluate = predicates_to_evaluate.copy()
        unresolved_predicates = []
        while predicates_to_evaluate:
            predicate = predicates_to_evaluate.pop(0)

            new_substitutions = []
            for sub in substitutions:
                new_substitutions += self.unify_builtin_substitution(
                    predicate, sub
                )

            if len(new_substitutions) == 0:
                unresolved_predicates.append(predicate)
            else:
                substitutions = self.compose_substitutions_no_conflict(
                    substitutions, new_substitutions
                )
                predicates_to_evaluate += unresolved_predicates
                unresolved_predicates = []

        if len(unresolved_predicates) == 0:
            return substitutions
        else:
            return []

    @staticmethod
    def compose_substitutions_no_conflict(substitutions, new_substitutions):
        composed_substitutions = []
        for substitution in substitutions:
            subs_keys = set(
                k for k, v in substitution.items()
                if not isinstance(v, Symbol)
            )
            composed_substitutions += ChaseGeneral.compose_all_subtitutions(
                new_substitutions,
                subs_keys,
                substitution
            )
        substitutions = composed_substitutions
        return substitutions

    @staticmethod
    def compose_all_subtitutions(new_substitutions, subs_keys, substitution):
        composed_subtitutions_ = []
        for new_substitution in new_substitutions:
            overlap = subs_keys & set(new_substitution)
            if any(
                not isinstance(new_substitution[k], Symbol) and
                new_substitution[k] != substitution[k]
                for k in overlap
            ):
                continue
            subs = compose_substitutions(
                substitution, new_substitution
            )
            composed_subtitutions_.append(subs)
        return composed_subtitutions_

    @staticmethod
    def compose_substitutions_ignoring_conflicts(
        new_substitutions,
        subs_keys,
        substitution
    ):
        for new_substitution in new_substitutions:
            overlap = subs_keys & set(new_substitution)
            if any(
                not isinstance(new_substitution[k], Symbol) and
                new_substitution[k] != substitution[k]
                for k in overlap
            ):
                continue
            subs = compose_substitutions(
                substitution, new_substitution
            )
        return subs

    def unify_builtin_substitution(self, predicate, substitution):
        substituted_predicate = apply_substitution(predicate, substitution)
        evaluated_predicate = self.datalog_program.walk(substituted_predicate)
        if (
            isinstance(evaluated_predicate, Constant[bool]) and
            evaluated_predicate.value
        ):
            return [substitution]
        elif self.is_equality_between_constant_and_symbol(evaluated_predicate):
            return [
                self.unify_builtin_substitution_equality(
                    evaluated_predicate
                )
            ]
        elif self.is_containment_of_symbol_in_constant(evaluated_predicate):
            return self.unify_builtin_substitution_containment(
                evaluated_predicate
            )
        else:
            return []

    @staticmethod
    def is_equality_between_constant_and_symbol(predicate):
        return (
            isinstance(predicate, FunctionApplication) and
            isinstance(predicate.functor, Constant) and
            predicate.functor.value is eq and
            any(isinstance(arg, Constant) for arg in predicate.args) and
            any(isinstance(arg, Symbol) for arg in predicate.args)
        )

    @staticmethod
    def is_containment_of_symbol_in_constant(predicate):
        return (
            isinstance(predicate, FunctionApplication) and
            isinstance(predicate.functor, Constant) and
            predicate.functor.value is contains and
            isinstance(predicate.args[0], Constant[Iterable]) and
            isinstance(predicate.args[1], Symbol)
        )

    @staticmethod
    def unify_builtin_substitution_equality(evaluated_predicate):
        if isinstance(evaluated_predicate.args[0], Symbol):
            substitution = {
                evaluated_predicate.args[0]: evaluated_predicate.args[1]
            }
        else:
            substitution = {
                evaluated_predicate.args[1]: evaluated_predicate.args[0]
            }
        return substitution

    @staticmethod
    def unify_builtin_substitution_containment(evaluated_predicate):
        set_value = evaluated_predicate.args[0].value
        symbol = evaluated_predicate.args[1]
        if len(set_value) == 0:
            return []
        elif isinstance(next(iter(set_value)), Constant):
            return [
                {symbol: v}
                for v in set_value
            ]
        else:
            el_type = ChaseGeneral.infer_iterable_subtype(evaluated_predicate)
            return [
                {
                    symbol:
                    Constant[el_type](
                        v,
                        auto_infer_type=False, verify_type=False
                    )
                }
                for v in set_value
            ]

    @staticmethod
    def infer_iterable_subtype(evaluated_predicate):
        iterable_subtype = evaluated_predicate.args[0].type
        type_args = get_args(iterable_subtype)
        el_type = type_args[0]
        if is_leq_informative(iterable_subtype, Tuple):
            for another_type in type_args[1:]:
                try:
                    el_type = unify_types(el_type, another_type)
                except NeuroLangTypeException:
                    el_type = Unknown
                    break
        return el_type

    def extract_rule_predicates(
        self, rule, instance, restriction_instance=None
    ):
        if restriction_instance is None:
            restriction_instance = dict()

        rule_predicates = extract_logic_predicates(rule.antecedent)
        restricted_predicates = []
        nonrestricted_predicates = []
        builtin_predicates = []
        for predicate in rule_predicates:
            functor = predicate.functor
            if functor in restriction_instance:
                restricted_predicates.append(
                    (predicate, restriction_instance[functor].value)
                )
            elif functor in instance:
                nonrestricted_predicates.append(
                    (predicate, instance[functor].value)
                )
            elif functor in self.builtins:
                builtin_predicates.append((predicate, self.builtins[functor]))
            elif isinstance(functor, Constant):
                builtin_predicates.append((predicate, functor))
            else:
                return ([], [], [])

        return (
            restricted_predicates, nonrestricted_predicates, builtin_predicates
        )

    def compute_result_set(
        self, rule, substitutions, instance, restriction_instance=None
    ):
        if restriction_instance is None:
            restriction_instance = MapInstance()

        row_type = None
        tuples = []
        for substitution in substitutions:
            tuple_, tuple_type = self.compute_new_tuple_and_type(
                rule,
                substitution
            )
            tuples.append(tuple_)
            row_type = self.aggregate_tuple_type_into_row_type(
                row_type,
                tuple_type
            )
        new_tuples = self.datalog_program.new_set(tuples, row_type=row_type)
        return self.compute_instance_update(
            rule, new_tuples, instance, restriction_instance
        )

    def aggregate_tuple_type_into_row_type(self, row_type, tuple_type):
        if row_type is None:
            row_type = tuple_type
        elif row_type is not tuple_type:
            row_type = unify_types(row_type, tuple_type)
        return row_type

    def compute_new_tuple_and_type(self, rule, substitution):
        tuple_ = tuple()
        type_ = tuple()
        for el in apply_substitution_arguments(
            rule.consequent.args, substitution
        ):
            tuple_ += (el.value,)
            type_ += (el.type,)
        tuple_type = Tuple[type_]
        return tuple_, tuple_type

    def compute_instance_update(
        self, rule, new_tuples, instance, restriction_instance
    ):
        instance_update = MapInstance({rule.consequent.functor: new_tuples})
        instance_update -= instance
        instance_update -= restriction_instance
        return instance_update

    def build_chase_tree(self, chase_set=chase_step):
        root = ChaseNode(
            MapInstance(self.datalog_program.extensional_database()),
            dict()
        )
        rules = []
        for union in self.datalog_program.intensional_database(
        ).values():
            for rule in union.formulas:
                rules.append(rule)

        nodes_to_process = [root]
        while len(nodes_to_process) > 0:
            node = nodes_to_process.pop(0)
            for rule in rules:
                new_node = self.build_nodes_from_rules(node, rule)
                if new_node is not None:
                    nodes_to_process.append(new_node)
        return root

    def build_nodes_from_rules(self, node, rule):
        instance_update = self.chase_step(node.instance, rule)
        if len(instance_update) > 0:
            new_instance = node.instance | instance_update
            new_node = ChaseNode(new_instance, dict())
            node.children[rule] = new_node
            return new_node
        else:
            return None

    def eliminate_already_computed(self, consequent, instance, substitutions):
        return substitutions

    def build_chase_solution(self):
        instance = MapInstance()
        instance_update = MapInstance(
            self.datalog_program.extensional_database()
        )
        self.check_constraints(instance_update)
        instance = self.execute_chase(self.rules, instance_update, instance)
        return instance


class ChaseNonRecursive:
    """Chase class for non-recursive programs.
    """

    def execute_chase(self, rules, instance_update, instance):
        rules_seen = set()
        while rules:
            rule = rules.pop(0)
            functor = rule.consequent.functor
            functor_ix = self._dependency_matrix_symbols.index(functor)
            if any(
                self._dependency_matrix_symbols[dep_index] not in rules_seen
                for dep_index in
                self._dependency_matrix[functor_ix].nonzero()[0]
            ):
                rules.append(rule)
                continue
            rules_seen.add(functor)

            with log_performance(LOG, 'Evaluating rule %s', (rule,)):
                instance_update |= self.chase_step(
                    instance, rule, restriction_instance=instance_update
                )
        return instance_update

    def check_constraints(self, instance_update):
        super().check_constraints(instance_update)
        self._dependency_matrix_symbols, self._dependency_matrix = \
            dependency_matrix(
                self.datalog_program, rules=self.rules
            )
        if program_has_loops(self._dependency_matrix):
            raise NeuroLangProgramHasLoopsException(
                "Use a different resolution algorithm"
            )


class ChaseNaive:
    """Chase implementation using the naive algorithm.
    """
    def execute_chase(self, rules, instance_update, instance):
        while len(instance_update) > 0:
            instance |= instance_update
            new_update = MapInstance()
            for rule in rules:
                with log_performance(LOG, 'Evaluating rule %s', (rule,)):
                    upd = self.chase_step(
                        instance | instance_update, rule,
                    )
                new_update |= upd
            instance_update = new_update
        return instance


class ChaseSemiNaive:
    """Chase implementation using the semi-naive algorithm.
    This algorithm will not work if there are non-linear rules.
       """

    def execute_chase(self, rules, instance_update, instance):
        continue_chase = len(instance_update) > 0
        while continue_chase:
            instance |= instance_update
            instance_update = MapInstance()
            continue_chase = False
            for rule in rules:
                with log_performance(LOG, 'Evaluating rule %s', (rule,)):
                    instance_update = self.per_rule_update(
                        rule, instance, instance_update
                    )
                continue_chase |= len(instance_update) > 0
        return instance

    def per_rule_update(self, rule, instance, instance_update):
        new_instance_update = self.chase_step(
            instance, rule, restriction_instance=instance_update
        )
        if len(new_instance_update) > 0:
            instance_update |= new_instance_update
        return instance_update

    def check_constraints(self, instance_update):
        super().check_constraints(instance_update)
        for rule in self.rules:
            if not is_linear_rule(rule):
                raise NeuroLangNonLinearProgramException(
                    f"Rule {rule} is non-linear. "
                    "Use a different resolution algorithm"
                )



from collections import namedtuple
from inspect import isclass
from itertools import tee
from functools import lru_cache
from typing import Tuple

import numpy as np

from ..expression_walker import ReplaceExpressionsByValues
from ..expressions import Constant
from ..type_system import (Unknown, get_args, infer_type,
                           unify_types)
from ..utils.relational_algebra_set import (
    NamedRelationalAlgebraFrozenSet, RelationalAlgebraFrozenSet,
    RelationalAlgebraSet)

REBV = ReplaceExpressionsByValues(dict())


class WrappedTypeMap:
    row_maps = {
        np.integer: int,
        np.float: float
    }

    @lru_cache(maxsize=256)
    def backend_2_python(self, value):
        for k, v in self.row_maps.items():
            if (
                (isclass(value) and issubclass(value, k)) or
                value == k
            ):
                return v
        return value


TYPEMAP = WrappedTypeMap()


class WrappedRelationalAlgebraSetBaseMixin:
    def __init__(
        self, iterable=None, row_type=Unknown, verify_row_type=True, **kwargs
    ):
        iterable = WrappedRelationalAlgebraSetBaseMixin._get_init_iterable(
            iterable
        )
        super().__init__(iterable=iterable, **kwargs)
        self._set_row_type(iterable, row_type, verify_row_type)

    def _set_row_type(self, iterable, row_type, verify_row_type):
        if row_type is not Unknown:
            if verify_row_type:
                raise NotImplementedError()
            self._row_type = row_type
        elif isinstance(iterable, WrappedRelationalAlgebraSetBaseMixin):
            self._row_type = iterable._row_type
        else:
            self._row_type = None

    @staticmethod
    def _get_init_iterable(iterable):
        if iterable is not None:
            if isinstance(
                iterable,
                (
                    WrappedRelationalAlgebraSetBaseMixin,
                    RelationalAlgebraFrozenSet
                )
            ):
                iterable = iterable
            elif hasattr(iterable, '__getitem__'):
                iterable = (
                    WrappedRelationalAlgebraSetBaseMixin
                    ._obtain_value_collection(
                        iterable
                    )
                )
            else:
                iterable = (
                    WrappedRelationalAlgebraSetBaseMixin
                    ._obtain_value_iterable(
                        iterable
                    )
                )
        return iterable

    def __contains__(self, element):
        element = REBV.walk(element)
        return super().__contains__(element)

    def _operator_wrapped(self, op, other):
        other_is_wras = isinstance(other, WrappedRelationalAlgebraSetBaseMixin)
        if other_is_wras:
            other = other.unwrap()
        else:
            other = {el for el in self._obtain_value_iterable(other)}
        operator = getattr(self.unwrap(), op)
        res = operator(other)
        if isinstance(res, WrappedRelationalAlgebraSetBaseMixin):
            res._row_type = self._get_new_row_type(other, other_is_wras)
        elif isinstance(res, RelationalAlgebraFrozenSet):
            res = type(self)(iterable=res)
        return res

    def _get_new_row_type(self, other, other_is_wras):
        row_type = self._row_type
        if other_is_wras:
            if row_type is not None and other._row_type is not None:
                row_type = unify_types(row_type, other._row_type)
            elif row_type is None:
                row_type = other._row_type
        return row_type

    @staticmethod
    def _obtain_value_iterable(iterable):
        it1, it2 = tee(iterable)
        iterator_of_constants = False
        for val in it1:
            iterator_of_constants = (
                WrappedRelationalAlgebraSetBaseMixin.
                is_constant_tuple_or_tuple_of_constants(val)
            )
            break
        if not iterator_of_constants:
            iterator = it2
        else:
            iterator = (REBV.walk(e) for e in it2)
        for e in iterator:
            yield e

    @staticmethod
    def _obtain_value_collection(iterable):
        if len(iterable) == 0:
            return iterable

        val = iterable[0]
        collection_of_constants = (
            WrappedRelationalAlgebraSetBaseMixin.
            is_constant_tuple_or_tuple_of_constants(val)
        )
        if not collection_of_constants:
            return iterable
        else:
            return (REBV.walk(e) for e in iterable)

    @staticmethod
    def is_constant_tuple_or_tuple_of_constants(val):
        return (
            isinstance(val, Constant[Tuple]) or
            (
                isinstance(val, tuple) and (len(val) > 0)
                and isinstance(val[0], Constant)
            )
        )

    def __eq__(self, other):
        return self._operator_wrapped('__eq__', other)

    def __ne__(self, other):
        return self._operator_wrapped('__ne__', other)

    def __lt__(self, other):
        return self._operator_wrapped('__lt__', other)

    def __gt__(self, other):
        return self._operator_wrapped('__gt__', other)

    def __le__(self, other):
        return self._operator_wrapped('__le__', other)

    def __ge__(self, other):
        return self._operator_wrapped('__ge__', other)

    def __and__(self, other):
        return self._operator_wrapped('__and__', other)

    def __or__(self, other):
        return self._operator_wrapped('__or__', other)

    def __sub__(self, other):
        return self._operator_wrapped('__sub__', other)

    def __hash__(self):
        return super().__hash__()

    def unwrapped_iter(self):
        return super().__iter__()

    @property
    def row_type(self):
        if self._row_type is None:
            if self.arity > 0 and not self.is_empty():
                self._row_type = Tuple[tuple(
                    TYPEMAP.backend_2_python(t)
                    for t in get_args(infer_type(super().fetch_one()))
                )]
            else:
                self._row_type = Tuple

        return self._row_type


class WrappedRelationalAlgebraFrozenSetMixin(
    WrappedRelationalAlgebraSetBaseMixin
):
    def __iter__(self):
        type_ = self.row_type
        element_types = get_args(type_)

        for t in super().__iter__():
            yield Constant[type_](
                tuple(
                    Constant[e_t](e, verify_type=False)
                    for e_t, e in zip(element_types, t)
                ),
                verify_type=False
            )


class WrappedRelationalAlgebraSetMixin(
    WrappedRelationalAlgebraSetBaseMixin
):
    def __iter__(self):
        type_ = self.row_type
        element_types = get_args(type_)

        for t in super().__iter__():
            yield Constant[type_](
                tuple(
                    Constant[e_t](e, verify_type=False)
                    for e_t, e in zip(element_types, t)
                ),
                verify_type=False
            )

    def add(self, value):
        return super().add(REBV.walk(value))

    def discard(self, value):
        return super().discard(REBV.walk(value))


def named_tuple_as_dict(*args, **kwargs):
    nt = namedtuple(*args, **kwargs)
    nt.get = lambda self, key, default=None: self._asdict().get(key, default)
    nt.keys = lambda self: self._asdict().keys()
    nt.items = lambda self: self._asdict().items()
    nt.values = lambda self: self._asdict().values()
    nt.__getitem__ = lambda self, key: self._asdict()[key]
    return nt


class WrappedNamedRelationalAlgebraFrozenSetMixin(
    WrappedRelationalAlgebraSetBaseMixin
):
    def __init__(
        self, columns=None, iterable=None,
        row_type=Unknown, verify_row_type=True, **kwargs
    ):
        iterable = WrappedRelationalAlgebraSetBaseMixin._get_init_iterable(
            iterable
        )
        if columns is None and iterable is not None:
            columns = iterable.columns
        super().__init__(columns=columns, iterable=iterable, **kwargs)
        self._set_row_type(iterable, row_type, verify_row_type)
        self.named_tuple_type = None

    @property
    def row_type(self):
        if self._row_type is None:
            if (self.arity > 0 and not self.is_empty()):
                element = super().fetch_one()
                self._row_type = Tuple[tuple(
                    Constant(getattr(element, c)).type
                    for c in self.columns
                )]
            else:
                self._row_type = Tuple

        return self._row_type

    def __iter__(self):
        if self.arity > 0:
            if self.named_tuple_type is None:
                self.named_tuple_type = named_tuple_as_dict(
                    'tuple', self.columns
                )

            row_types = {
                c: t
                for c, t in zip(self.columns, get_args(self.row_type))
            }
            for row in super().__iter__():
                nt = self.named_tuple_type(**{
                    f: Constant[row_types[f]](
                        v, verify_type=False
                    )
                    for f, v in zip(row._fields, row)
                })
                yield nt
        else:
            for _ in range(len(self)):
                yield dict()


class WrappedRelationalAlgebraFrozenSet(
    WrappedRelationalAlgebraFrozenSetMixin,
    RelationalAlgebraFrozenSet
):
    def unwrap(self):
        return RelationalAlgebraFrozenSet.create_view_from(self)


class WrappedRelationalAlgebraSet(
    WrappedRelationalAlgebraSetMixin,
    RelationalAlgebraSet
):
    def unwrap(self):
        return RelationalAlgebraSet.create_view_from(self)


class WrappedNamedRelationalAlgebraFrozenSet(
    WrappedNamedRelationalAlgebraFrozenSetMixin,
    NamedRelationalAlgebraFrozenSet
):
    def unwrap(self):
        return NamedRelationalAlgebraFrozenSet.create_view_from(self)



import io
from typing import AbstractSet, Tuple

import numpy as np
import pytest

from ...exceptions import UnsupportedProgramError, UnsupportedQueryError
from ...probabilistic.exceptions import (
    ForbiddenConditionalQueryNonConjunctive,
    UnsupportedProbabilisticQueryError
)
from ...utils.relational_algebra_set import RelationalAlgebraFrozenSet
from ..probabilistic_frontend import ProbabilisticFrontend


def assert_almost_equal(set_a, set_b):
    for tupl_a in set_a:
        assert any(
            all(
                np.isclose(term_a, term_b)
                if isinstance(term_a, (float, np.float32, np.float64))
                else term_a == term_b
                for term_a, term_b in zip(tupl_a, tupl_b)
            )
            for tupl_b in set_b
        )


def test_add_uniform_probabilistic_choice_set():
    nl = ProbabilisticFrontend()

    prob = [(a,) for a in range(10)]
    prob_set = nl.add_uniform_probabilistic_choice_over_set(prob, "prob")
    res = nl[prob_set]

    assert prob_set.type is AbstractSet[Tuple[float, int]]
    assert res.type is AbstractSet[Tuple[float, int]]
    assert res.value == frozenset((1 / len(prob), a) for a in range(10))

    d = [("a",), ("b",), ("c",), ("d",)]
    data = nl.add_uniform_probabilistic_choice_over_set(d, name="data")
    res_d = nl[data]

    assert data.type is AbstractSet[Tuple[float, str]]
    assert res_d.type is AbstractSet[Tuple[float, str]]


def test_deterministic_query():
    nl = ProbabilisticFrontend()
    d1 = [(1,), (2,), (3,), (4,), (5,)]
    data1 = nl.add_tuple_set(d1, name="data1")

    d2 = [(2, "a"), (3, "b"), (4, "d"), (5, "c"), (7, "z")]
    data2 = nl.add_tuple_set(d2, name="data2")

    with nl.scope as e:
        e.query1[e.y] = data1[e.x] & data2[e.x, e.y]
        res = nl.solve_all()

    assert "query1" in res.keys()
    q1 = res["query1"].as_pandas_dataframe().values
    assert len(q1) == 4
    for elem in q1:
        assert elem[0] in ["a", "b", "c", "d"]


def test_probabilistic_query():
    nl = ProbabilisticFrontend()
    d1 = [(1,), (2,), (3,), (4,), (5,)]
    data1 = nl.add_uniform_probabilistic_choice_over_set(d1, name="data1")

    d2 = [(2, "a"), (3, "b"), (4, "d"), (5, "c"), (7, "z")]
    data2 = nl.add_uniform_probabilistic_choice_over_set(d2, name="data2")

    d3 = [("a",), ("b",), ("c",)]
    data3 = nl.add_uniform_probabilistic_choice_over_set(d3, name="data3")

    with nl.scope as e:
        e.query1[e.y, e.PROB[e.y]] = data1[e.x] & data2[e.x, e.y]
        e.query2[e.y, e.PROB[e.y]] = e.query1[e.y] & data3[e.y]
        with pytest.raises(UnsupportedProgramError):
            nl.solve_all()


def test_marg_query():
    nl = ProbabilisticFrontend()
    nl.add_probabilistic_choice_from_tuples(
        {(0.2, "a"), (0.3, "b"), (0.5, "c")}, name="P"
    )
    nl.add_tuple_set({(1, "a"), (2, "a"), (2, "b")}, name="Q")
    nl.add_tuple_set({(1, "a"), (1, "b"), (2, "b"), (2, "c")}, name="R")

    with nl.scope as e:
        e.Z[e.x, e.z, e.PROB[e.x, e.z]] = (e.Q(e.x, e.y) & e.P(e.y)) // (
            e.R(e.z, e.y) & e.P(e.y)
        )

        res = nl.solve_all()

    expected = RelationalAlgebraFrozenSet(
        {
            (1, 1, 0.4),
            (2, 1, 1.0),
            (2, 2, 0.375),
        }
    )
    assert_almost_equal(res["Z"], expected)


def test_mixed_queries():
    nl = ProbabilisticFrontend()
    d1 = [(1,), (2,), (3,), (4,), (5,)]
    data1 = nl.add_tuple_set(d1, name="data1")

    d2 = [(2, "a"), (3, "b"), (4, "d"), (5, "c"), (7, "z")]
    data2 = nl.add_tuple_set(d2, name="data2")

    d3 = [("a",), ("b",), ("c",), ("d",)]
    data3 = nl.add_uniform_probabilistic_choice_over_set(d3, name="data3")

    with nl.scope as e:
        e.tmp[e.x, e.y] = data1[e.x] & data2[e.x, e.y]
        e.query1[e.x, e.y, e.PROB[e.x, e.y]] = e.tmp[e.x, e.y]
        e.query2[e.y, e.PROB[e.y]] = e.tmp[e.x, e.y] & data3[e.y]
        res = nl.solve_all()

    assert "query1" in res.keys()
    assert "query2" in res.keys()
    assert len(res["query2"].columns) == 2
    q2 = res["query2"].as_pandas_dataframe().values
    assert len(q2) == 4
    for elem in q2:
        assert elem[1] == 0.25
        assert elem[0] in ["a", "b", "c", "d"]


def test_ontology_query():

    test_case = """
    <rdf:RDF
        xmlns:rdf="http://www.w3.org/1999/02/22-rdf-syntax-ns#"
        xmlns:rdfs="http://www.w3.org/2000/01/rdf-schema#"
        xmlns:owl="http://www.w3.org/2002/07/owl#"
        xmlns:first="http://www.w3.org/2002/03owlt/hasValue/premises001#"
        xml:base="http://www.w3.org/2002/03owlt/hasValue/premises001" >
        <owl:Ontology/>
        <owl:Class rdf:ID="r">
        <rdfs:subClassOf>
            <owl:Restriction>
                <owl:onProperty rdf:resource="#p2"/>
                <owl:hasValue rdf:datatype="http://www.w3.org/2001/XMLSchema#boolean">true</owl:hasValue>
            </owl:Restriction>
        </rdfs:subClassOf>
        </owl:Class>
        <owl:ObjectProperty rdf:ID="p"/>
        <owl:ObjectProperty rdf:ID="p2"/>
        <owl:Class rdf:ID="c"/>
        <first:r rdf:ID="i">
        <rdf:type rdf:resource="http://www.w3.org/2002/07/owl#Thing"/>
        <first:p>
            <owl:Thing rdf:ID="o" />
        </first:p>
        </first:r>
    </rdf:RDF>
    """

    nl = ProbabilisticFrontend()
    nl.load_ontology(io.StringIO(test_case))

    p2 = nl.new_symbol(
        name="http://www.w3.org/2002/03owlt/hasValue/premises001#p2"
    )

    with nl.scope as e:
        e.answer[e.x, e.y] = p2[e.x, e.y]
        solution_instance = nl.solve_all()

    resp = list(solution_instance["answer"].itervalues())
    assert (
        "http://www.w3.org/2002/03owlt/hasValue/premises001#i",
        "true",
    ) in resp


def test_simple_within_language_succ_query():
    nl = ProbabilisticFrontend()
    P = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("b",), ("c",)], name="P"
    )
    Q = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("d",), ("e",)], name="Q"
    )
    with nl.scope as e:
        e.Z[e.x, e.PROB[e.x]] = P[e.x] & Q[e.x]
        res = nl.solve_all()
    assert "Z" in res.keys()
    df = res["Z"].as_pandas_dataframe()
    assert len(df) == 1
    assert tuple(df.values[0])[0] == "a"
    assert np.isclose(tuple(df.values[0])[1], 1 / 9)


def test_within_language_succ_query():
    nl = ProbabilisticFrontend()
    P = nl.add_uniform_probabilistic_choice_over_set(
        [
            ("a", "b"),
            ("b", "c"),
            ("b", "d"),
        ],
        name="P",
    )
    Q = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("b",)], name="Q"
    )
    with nl.scope as e:
        e.Z[e.x, e.PROB[e.x]] = P[e.x, e.y] & Q[e.x]
        res = nl.solve_all()
    assert "Z" in res.keys()
    df = res["Z"].as_pandas_dataframe()
    assert np.isclose(df.loc[df.iloc[:, 0] == "b"].iloc[0, 1], 2 / 3 / 2)
    assert np.isclose(df.loc[df.iloc[:, 0] == "a"].iloc[0, 1], 1 / 3 / 2)


def test_solve_query():
    nl = ProbabilisticFrontend()
    P = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("b",), ("c",)], name="P"
    )
    Q = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("d",), ("c",)], name="Q"
    )
    with nl.scope as e:
        e.Z[e.x, e.PROB[e.x]] = P[e.x] & Q[e.x]
        res = nl.query((e.x, e.p), e.Z[e.x, e.p])
    expected = RelationalAlgebraFrozenSet(
        [
            ("a", 1 / 9),
            ("c", 1 / 9),
        ]
    )
    assert_almost_equal(res, expected)


def test_solve_query_prob_col_not_last():
    nl = ProbabilisticFrontend()
    P = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("b",), ("c",)], name="P"
    )
    Q = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("d",), ("c",)], name="Q"
    )
    with nl.scope as e:
        e.Z[e.PROB[e.x], e.x] = P[e.x] & Q[e.x]
        res = nl.query((e.p, e.x), e.Z[e.p, e.x])
    expected = RelationalAlgebraFrozenSet(
        [
            (1 / 9, "a"),
            (1 / 9, "c"),
        ]
    )
    assert_almost_equal(res, expected)


def test_solve_boolean_query():
    nl = ProbabilisticFrontend()
    P = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("b",), ("c",)], name="P"
    )
    Q = nl.add_uniform_probabilistic_choice_over_set(
        [("a",), ("d",), ("c",)], name="Q"
    )
    with pytest.raises(UnsupportedProbabilisticQueryError):
        with nl.scope as e:
            e.ans[e.PROB()] = P[e.x] & Q[e.x]
            nl.query((e.p), e.ans[e.p])


def test_solve_complex_stratified_query():
    """
    R(1, 2) : 0.3
    R(1, 4) : 0.7
    R(2, 2) : 0.2
    R(2, 4) : 0.6
    R(2, 6) : 0.8
    Q(4) : 0.2 v Q(6) : 0.8
    A(x, PROB[x])       :- Q(x), R(1, x), R(2, x)
    B(x, y, PROB[x, y]) :- Q(y), R(1, x), R(2, x)
    C(x, y, p1, p2)     :- A(x, p1), B(x, y, p2)

    """
    nl = ProbabilisticFrontend()
    R = nl.add_probabilistic_facts_from_tuples(
        [(0.3, 1, 2), (0.7, 1, 4), (0.2, 2, 2), (0.6, 2, 4), (0.8, 2, 6)],
        name="R",
    )
    Q = nl.add_probabilistic_choice_from_tuples([(0.2, 4), (0.8, 6)], name="Q")
    with nl.scope as e:
        e.A[e.x, e.PROB[e.x]] = Q[e.x] & R[1, e.x] & R[2, e.x]
        e.B[e.x, e.y, e.PROB[e.x, e.y]] = Q[e.y] & R[1, e.x] & R[2, e.x]
        e.C[e.x, e.y, e.p1, e.p2] = e.A[e.x, e.p1] & e.B[e.x, e.y, e.p2]
        res = nl.query((e.x, e.y, e.h, e.z), e.C[e.x, e.y, e.h, e.z])
    expected = RelationalAlgebraFrozenSet(
        [
            (4, 4, 0.2 * 0.7 * 0.6, 0.2 * 0.7 * 0.6),
            (4, 6, 0.2 * 0.7 * 0.6, 0.8 * 0.7 * 0.6),
        ]
    )
    assert_almost_equal(res, expected)


def test_solve_complex_stratified_query_with_deterministic_part():
    nl = ProbabilisticFrontend()
    A = nl.add_tuple_set([("a",), ("b",), ("c",)], name="A")
    B = nl.add_tuple_set([("a",), ("b",)], name="B")
    P = nl.add_probabilistic_facts_from_tuples(
        [(0.2, "a"), (0.8, "b")],
        name="P",
    )
    with nl.scope as e:
        e.C[e.x, e.y] = A[e.x] & A[e.y] & B[e.x]
        e.D[e.x, e.y, e.PROB[e.x, e.y]] = e.C[e.x, e.y] & P[e.y]
        res = nl.query((e.x, e.y, e.p), e.D[e.x, e.y, e.p])
    expected = RelationalAlgebraFrozenSet(
        [
            ("a", "a", 0.2),
            ("a", "b", 0.8),
            ("b", "a", 0.2),
            ("b", "b", 0.8),
        ]
    )
    assert_almost_equal(res, expected)


def test_neurolange_dl_deterministic_negation():
    neurolang = ProbabilisticFrontend()
    s = neurolang.new_symbol(name="s")
    x = neurolang.new_symbol(name="x")
    y = neurolang.new_symbol(name="y")

    dataset = {(i, i * 2) for i in range(10)}
    q = neurolang.add_tuple_set(dataset, name="q")
    s[x, y] = ~q(x, x) & q(x, y)

    res = neurolang.solve_all()

    assert res["s"].to_unnamed() == {(i, j) for i, j in dataset if i != j}


@pytest.mark.xfail(reason="RAP negation needs to be implemented")
def test_neurolange_dl_probabilistic_negation():
    neurolang = ProbabilisticFrontend()
    s = neurolang.new_symbol(name="s")
    x = neurolang.new_symbol(name="x")
    y = neurolang.new_symbol(name="y")
    prob = neurolang.new_symbol(name="PROB")

    dataset_det = {(i, i * 2) for i in range(10)}
    dataset = {((1 + i) / 10, i, i * 2) for i in range(10)}
    q = neurolang.add_tuple_set(dataset_det, name="q")
    r = neurolang.add_probabilistic_facts_from_tuples(dataset, name="r")

    s[x, y, prob(x, y)] = ~r(x, x) & q(x, y)

    res = neurolang.solve_all()

    assert res["s"].to_unnamed() == {(i, j) for i, j in dataset if i != j}


def test_neurolang_dl_aggregation():
    neurolang = ProbabilisticFrontend()
    q = neurolang.new_symbol(name="q")
    p = neurolang.new_symbol(name="p")
    r = neurolang.new_symbol(name="r")
    x = neurolang.new_symbol(name="x")
    y = neurolang.new_symbol(name="y")

    @neurolang.add_symbol
    def sum_(x):
        return sum(x)

    for i in range(10):
        q[i % 2, i] = True

    p[x, sum_(y)] = q[x, y]

    sol = neurolang.query(r(x, y), p(x, y))

    res_q = {(0, 2 + 4 + 8), (1, 1 + 5 + 9)}

    assert len(sol) == 2
    assert sol[r] == res_q
    assert sol[p] == res_q


def test_post_probabilistic_aggregation():
    nl = ProbabilisticFrontend()
    A = nl.add_probabilistic_facts_from_tuples(
        [(0.2, "a"), (0.9, "b"), (0.5, "c")],
        name="A",
    )
    B = nl.add_probabilistic_choice_from_tuples(
        [(0.2, "a", "c"), (0.7, "b", "c"), (0.1, "a", "d")],
        name="B",
    )

    @nl.add_symbol
    def mysum(x):
        return sum(x)

    with nl.scope as e:
        e.C[e.x, e.y, e.PROB[e.x, e.y]] = A[e.x] & B[e.x, e.y]
        e.D[e.x, e.mysum(e.p)] = e.C[e.x, e.y, e.p]
        res = nl.query((e.x, e.s), e.D[e.x, e.s])

    assert len(res) == 2
    assert res == {("a", 0.2 * 0.2 + 0.2 * 0.1), ("b", 0.9 * 0.7)}


def test_empty_result_query():
    nl = ProbabilisticFrontend()
    A = nl.add_tuple_set([("f",), ("d",)], name="A")
    B = nl.add_probabilistic_facts_from_tuples(
        [
            (0.2, "a"),
            (0.7, "b"),
            (0.6, "c"),
        ],
        name="B",
    )
    with nl.scope as e:
        e.Q[e.x, e.PROB[e.x]] = A[e.x] & B[e.x]
        res = nl.query((e.x, e.p), e.Q[e.x, e.p])
    assert res.is_empty()
    with nl.scope as e:
        e.Q[e.x] = A[e.x] & A["b"]
        res = nl.query((e.x,), e.Q[e.x])
    assert res.is_empty()


def test_forbidden_query_on_probabilistic_predicate():
    nl = ProbabilisticFrontend()
    A = nl.add_tuple_set([("f",), ("d",)], name="A")
    B = nl.add_probabilistic_facts_from_tuples(
        [(0.2, "a"), (0.7, "b"), (0.6, "c")], name="B"
    )
    with pytest.raises(UnsupportedQueryError):
        with nl.scope as e:
            e.Q[e.x] = A[e.x] & B[e.x] & A["b"]
            nl.query((e.x,), e.Q[e.x])


def test_empty_boolean_query_result():
    nl = ProbabilisticFrontend()
    A = nl.add_tuple_set([("a",), ("b",), ("c",)], name="A")
    B = nl.add_probabilistic_facts_from_tuples(
        [(0.4, "a"), (0.5, "b")], name="B"
    )
    C = nl.add_probabilistic_choice_from_tuples(
        [(0.8, "c"), (0.2, "f")], name="C"
    )
    with nl.scope as e:
        e.D[e.x, e.PROB[e.x]] = A[e.x] & B[e.x] & C[e.x]
        res = nl.query(tuple(), e.D[e.x, e.p])
    assert not res


def test_equality():
    nl = ProbabilisticFrontend()
    r1 = nl.add_tuple_set([(i,) for i in range(5)], name="r1")

    with nl.scope as e:
        e.r2[e.x] = r1(e.y) & (e.x == e.y * 2)

        sol = nl.query((e.x,), e.r2[e.x])

    assert set(sol) == set((2 * i,) for i in range(5))

    r2 = nl.add_tuple_set([("Hola",), ("Hello",), ("Bonjour",)], name="r2")

    @nl.add_symbol
    def lower(input: str) -> str:
        return input.lower()

    with nl.scope as e:
        e.r3[e.x] = r2(e.y) & (e.x == lower(e.y))

        sol = nl.query((e.x,), e.r3[e.x])

    assert set(sol) == set((("hola",), ("hello",), ("bonjour",)))


def test_equality2():
    nl = ProbabilisticFrontend()
    nl.add_tuple_set(
        [("Hola", "var"), ("Hello", "var2"), ("Bonjour", "var")],
        name="test_var",
    )

    @nl.add_symbol
    def word_lower(name: str) -> str:
        return str(name).lower()

    with nl.scope as e:
        e.low[e.lower] = e.test_var[e.name, "var"] & (
            e.lower == word_lower(e.name)
        )

        query = nl.query((e.lower,), e.low[e.lower])

    assert set(query) == set((("hola",), ("bonjour",)))


def test_result_both_deterministic_and_post_probabilistic():
    nl = ProbabilisticFrontend()
    nl.add_tuple_set(
        [
            ("this", "is", "inglese"),
            ("questo", "è", "italiano"),
            ("ceci", "est", "francese"),
        ],
        name="parole",
    )
    nl.add_probabilistic_choice_from_tuples(
        [
            (0.4, "valentino"),
            (0.6, "pietro"),
        ],
        name="persona_scelta",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.1, "italiano"),
            (0.7, "inglese"),
            (0.2, "francese"),
        ],
        name="lingua_parlata",
    )
    with nl.scope as e:
        e.e_una_lingua[e.lingua] = e.parole[e.subject, e.verb, e.lingua]
        e.lingua_da_lua_decisa[
            e.persona, e.lingua, e.PROB[e.persona, e.lingua]
        ] = (
            e.e_una_lingua[e.lingua]
            & e.persona_scelta[e.persona]
            & e.lingua_parlata[e.lingua]
        )
        e.utilizzare_le_probabilita[e.lingua, e.p] = e.lingua_da_lua_decisa[
            ..., e.lingua, e.p
        ] & (e.p > 0.1)
        res = nl.solve_all()
    assert "e_una_lingua" in res
    assert "lingua_da_lua_decisa" in res
    assert "utilizzare_le_probabilita" in res
    assert len(res["utilizzare_le_probabilita"]) == 3
    expected = RelationalAlgebraFrozenSet(
        [
            ("francese", 0.12),
            ("inglese", 0.7 * 0.4),
            ("inglese", 0.7 * 0.6),
        ]
    )
    assert_almost_equal(res["utilizzare_le_probabilita"], expected)


def test_result_query_relation_correct_column_names():
    nl = ProbabilisticFrontend()
    nl.add_tuple_set(
        [
            ("alice",),
            ("bob",),
        ],
        name="person",
    )
    nl.add_tuple_set(
        [("alice", "paris"), ("bob", "marseille")],
        name="lives_in",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.8, "bob"),
            (0.9, "alice"),
        ],
        name="does_not_smoke",
    )
    with nl.environment as e:
        e.climbs[e.p, e.PROB[e.p, e.city], e.city] = (
            e.person[e.p] & e.lives_in[e.p, e.city] & e.does_not_smoke[e.p]
        )
        solution = nl.solve_all()
    assert all(name in solution for name in ["climbs", "person", "lives_in"])


def test_add_constraints_and_rewrite():
    nl = ProbabilisticFrontend()

    nl.add_tuple_set(
        [
            ("neurolang"),
        ],
        name="project",
    )

    nl.add_tuple_set(
        [
            ("neurolang", "db"),
            ("neurolang", "logic"),
        ],
        name="inArea",
    )

    with nl.scope as e:
        nl.add_constraint(
            e.project[e.x] & e.inArea[e.x, e.y],
            e.hasCollaborator[e.z, e.y, e.x],
        )

        e.p[e.b] = e.hasCollaborator[e.a, "db", e.b]
        res = nl.solve_all()

    assert list(res["p"])[0] == ("neurolang",)


def test_solve_marg_query():
    nl = ProbabilisticFrontend()
    nl.add_tuple_set(
        [
            ("alice",),
            ("bob",),
        ],
        name="person",
    )
    nl.add_tuple_set(
        [("alice", "paris"), ("bob", "marseille")],
        name="lives_in",
    )
    nl.add_probabilistic_choice_from_tuples(
        [
            (0.2, "alice", "running"),
            (0.8, "bob", "climbing"),
        ],
        name="practice",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.8, "bob"),
            (0.9, "alice"),
        ],
        name="does_not_smoke",
    )
    with nl.environment as e:
        e.query[e.p, e.PROB[e.p, e.city, e.sport], e.city, e.sport] = (
            e.person[e.p] & e.lives_in[e.p, e.city] & e.does_not_smoke[e.p]
        ) // e.practice[e.p, e.sport]
        solution = nl.solve_all()
    assert all(name in solution for name in ["query", "person", "lives_in"])
    with nl.environment as e:
        result = nl.query(
            (e.p, e.prob, e.city, e.sport),
            e.query(e.p, e.prob, e.city, e.sport),
        )
    expected = RelationalAlgebraFrozenSet(
        iterable=[
            ("alice", 0.9, "paris", "running"),
            ("bob", 0.8, "marseille", "climbing"),
        ],
    )
    assert_almost_equal(result, expected)


def test_solve_marg_query_disjunction():
    nl = ProbabilisticFrontend()
    nl.add_tuple_set(
        [
            ("alice",),
            ("bob",),
        ],
        name="person",
    )
    nl.add_tuple_set(
        [("alice", "paris"), ("bob", "marseille")],
        name="lives_in",
    )
    nl.add_probabilistic_choice_from_tuples(
        [
            (0.2, "alice", "running"),
            (0.8, "bob", "climbing"),
        ],
        name="practice",
    )
    nl.add_probabilistic_facts_from_tuples(
        [
            (0.8, "bob"),
            (0.9, "alice"),
        ],
        name="does_not_smoke",
    )
    with pytest.raises(ForbiddenConditionalQueryNonConjunctive):
        with nl.environment as e:
            e.query[e.p, e.PROB[e.p, e.city, e.sport], e.city, e.sport] = (
                e.person[e.p] & (
                    e.lives_in[e.p, e.city] |
                    e.does_not_smoke[e.p]
                )
            ) // e.practice[e.p, e.sport]



from abc import abstractmethod, abstractproperty
from collections.abc import MutableSet, Set


class RelationalAlgebraFrozenSet(Set):
    def __init__(self, columns):
        raise NotImplementedError()

    @classmethod
    def create_view_from(cls, other):
        raise NotImplementedError()

    @classmethod
    def dee(cls):
        raise NotImplementedError()

    @classmethod
    def dum(cls):
        raise NotImplementedError()

    @abstractmethod
    def is_empty(self):
        pass

    def is_dum(self):
        return (
            self.arity == 0 and
            self.is_empty()
        )

    def is_dee(self):
        return (
            self.arity == 0 and
            not self.is_empty()
        )

    @abstractmethod
    def __contains__(self, element):
        pass

    @abstractmethod
    def __iter__(self):
        pass

    @abstractmethod
    def fetch_one(self):
        pass

    @abstractproperty
    def arity(self):
        pass

    @abstractproperty
    def columns(self):
        pass

    @abstractmethod
    def projection(self, *columns):
        pass

    @abstractmethod
    def selection(self, select_criteria):
        pass

    @abstractmethod
    def selection_columns(self, select_criteria):
        pass

    @abstractmethod
    def equijoin(self, other, join_indices, return_mappings=False):
        pass

    @abstractmethod
    def cross_product(self, other):
        pass

    @abstractmethod
    def copy(self):
        pass

    @abstractmethod
    def groupby(self, columns):
        pass

    @abstractmethod
    def itervalues(self):
        pass

    @abstractmethod
    def as_numpy_array():
        pass

    def _dee_dum_product(self, other):
        res = None
        if self.is_dum():
            res = self
        elif other.is_dum():
            res = other
        elif self.is_dee():
            res = other
        elif other.is_dee():
            res = self
        return res

    def _dee_dum_sum(self, other):
        res = None
        if self.is_dum():
            res = other
        elif other.is_dum():
            res = self
        elif self.is_dee() and other.is_dee():
            return self
        return res


class NamedRelationalAlgebraFrozenSet(RelationalAlgebraFrozenSet):
    def __init__(self, columns, iterable=None):
        raise NotImplementedError()

    @classmethod
    def dee(cls):
        raise NotImplementedError()

    @classmethod
    def dum(cls):
        raise NotImplementedError()

    @abstractproperty
    def columns(self):
        pass

    @abstractproperty
    def arity(self):
        pass

    @abstractmethod
    def __contains__(self, element):
        pass

    @abstractmethod
    def projection(self, *columns):
        pass

    def equijoin(self, other, join_indices, return_mappings=False):
        raise NotImplementedError()

    @abstractmethod
    def naturaljoin(self, other):
        pass

    @abstractmethod
    def cross_product(self, other):
        pass

    @abstractmethod
    def rename_column(self, src, dst):
        pass

    @abstractmethod
    def rename_columns(self, renames):
        pass

    @abstractmethod
    def groupby(self, columns):
        pass

    @abstractmethod
    def aggregate(self, group_columns, aggregate_function):
        pass

    @abstractmethod
    def extended_projection(self, eval_expressions):
        pass

    @abstractmethod
    def __iter__(self):
        pass

    @abstractmethod
    def fetch_one(self):
        pass

    @abstractmethod
    def to_unnamed(self):
        pass

    def __lt__(self, other):
        raise NotImplementedError()

    def __le__(self, other):
        raise NotImplementedError()

    def __gt__(self, other):
        raise NotImplementedError()

    def __ge__(self, other):
        raise NotImplementedError()


class RelationalAlgebraSet(RelationalAlgebraFrozenSet, MutableSet):

    @abstractmethod
    def add(self, value):
        pass

    @abstractmethod
    def discard(self, value):
        pass



import problog.core
import problog.logic
import problog.sdd_formula

from ....datalog.expressions import Fact
from ....expressions import Constant, Symbol
from ....logic import Implication
from ..problog_solver import cplogic_to_problog
from ..program import CPLogicProgram

P = Symbol("P")
a = Constant("a")


def test_convert_cpl_to_pl():
    cpl_program = CPLogicProgram()
    cpl_program.add_probabilistic_facts_from_tuples(
        P, {(0.2, "a"), (1.0, "b")}
    )
    pl = cplogic_to_problog(cpl_program)
    query = problog.logic.Term("query")
    query_pred = problog.logic.Term("P", problog.logic.Var("v"))
    pl += query(query_pred)
    res = problog.core.ProbLog.convert(pl, problog.sdd_formula.SDD).evaluate()
    expected = {
        problog.logic.Term("P")(problog.logic.Constant("a")): 0.2,
        problog.logic.Term("P")(problog.logic.Constant("b")): 1.0,
    }
    assert res == expected


def test_zero_arity():
    cpl_program = CPLogicProgram()
    cpl_program.walk(Fact(P()))
    cplogic_to_problog(cpl_program)
    cpl_program = CPLogicProgram()
    cpl_program.walk(Fact(P(a)))
    cpl_program.walk(Implication(Symbol("yes")(), P(a)))
    pl = cplogic_to_problog(cpl_program)
    query = problog.logic.Term("query")
    query_pred = problog.logic.Term("yes")
    pl += query(query_pred)
    res = problog.core.ProbLog.convert(pl, problog.sdd_formula.SDD).evaluate()
    expected = {problog.logic.Term("yes"): 1.0}
    assert res == expected



r"""
Query Builder Base, Region and Neurosynth Mixins
================================================
Base classes to construct Datalog programs.
Capabilities to declare, manage and manipulate symbols.
Mixins provide capabilities related to brain volumes and
Neurosynth metadata manipulation
"""
from contextlib import contextmanager
from typing import (
    AbstractSet,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Tuple,
    Type,
    Union,
)
from uuid import uuid1

import numpy as np
from nibabel.dataobj_images import DataobjImage

from .. import expressions as ir
from ..region_solver import Region
from ..regions import ExplicitVBR, ImplicitVBR, SphericalVolume
from ..type_system import Unknown, is_leq_informative
from ..typed_symbol_table import TypedSymbolTable
from .neurosynth_utils import NeuroSynthHandler, StudyID, TfIDf
from . import query_resolution_expressions as fe
from ..datalog import DatalogProgram
from ..logic import ExistentialPredicate, UniversalPredicate


class QueryBuilderBase:
    """
    Base class to construct Datalog programs.
    Capabilities to declare, manage and manipulate symbols.
    """

    def __init__(
        self, program_ir: DatalogProgram, logic_programming: bool = False
    ) -> "QueryBuilderBase":
        """
        Query Builder with symbol management capabilities

        Parameters
        ----------
        program_ir : DatalogProgram
            Datalog program's intermediate representation,
            usually blank
        logic_programming : bool, optional
            defines if symbols can be dynamically
            declared, by default False

        Returns
        -------
        QueryBuilderBase
            see description
        """
        self.program_ir = program_ir
        self.set_type = AbstractSet[self.program_ir.type]
        self.logic_programming = logic_programming

        for k, v in self.program_ir.included_functions.items():
            self.symbol_table[ir.Symbol[v.type](k)] = v

        self._symbols_proxy = QuerySymbolsProxy(self)

    def get_symbol(self, symbol_name: Union[str, fe.Expression]) -> fe.Symbol:
        """
        Retrieves symbol via its name, either providing a
        fe.Expression with the correct name or the name itself

        Parameters
        ----------
        symbol_name : Union[str, fe.Expression]
            name of the symbol to be retrieved. If of type fe.Expression,
            expression's name is used as the name

        Returns
        -------
        fe.Symbol
            symbol corresponding to the input name

        Raises
        ------
        ValueError
            if no symbol could be found with given name

        Example
        -------
        >>> p_ir = DatalogProgram()
        >>> nl = QueryBuilderBase(program_ir=p_ir)
        >>> nl.add_symbol(3, "x")
        >>> nl.get_symbol("x")
        x: <class 'int'> = 3
        """
        if isinstance(symbol_name, fe.Expression):
            symbol_name = symbol_name.expression.name
        if symbol_name not in self.symbol_table:
            raise ValueError(f"fe.Symbol {symbol_name} not defined")
        return fe.Symbol(self, symbol_name)

    def __getitem__(
        self, symbol_name: Union[fe.Symbol, str, fe.Expression]
    ) -> fe.Symbol:
        """
        Retrieves symbol via its name, either providing a
        fe.Expression with the correct name or the name itself
        Points towards .get_symbol method

        Parameters
        ----------
        symbol_name : Union[fe.Symbol, str, fe.Expression]
            if of type fe.Symbol, symol's name will be passed
            to the .get_symbol method

        Returns
        -------
        fe.Symbol
            output from .get_symbol method
        """
        if isinstance(symbol_name, fe.Symbol):
            symbol_name = symbol_name.symbol_name
        return self.get_symbol(symbol_name)

    def __contains__(self, symbol: fe.Symbol) -> bool:
        """
        Checks if symbol exists in current symbol_table

        Parameters
        ----------
        symbol : fe.Symbol
            symbol to check

        Returns
        -------
        bool
            does symbol exist in current symbol_table
        """
        return symbol in self.symbol_table

    def exists(
        self, quantified_variable: fe.Symbol, body: fe.Expression
    ) -> fe.Exists:
        """
        Existential predicate on the body. The predicate
        will be cosidered satisfied if any instance of
        `quantified_variable` satifies `body`.

        Parameters
        ----------
        quantified_variable : fe.Symbol
            existentially-quantified variable

        body: fe.Expression
            first order logic proposition which to be
            satisfied by the existential quatifier.

        Returns
        -------
        fe.Exists
            representation of the existential predicate.
        """
        existential_predicate = (
            fe.Exists(
                self,
                ExistentialPredicate(
                    quantified_variable.expression,
                    body.expression
                ),
                quantified_variable,
                body
            )
        )

        return existential_predicate

    def all(
        self, quantified_variable: fe.Symbol, body: fe.Expression
    ) -> fe.All:
        """
        Universal predicate on the body. The predicate
        will be cosidered satisfied if all instances of
        `quantified_variable` satify `body`.

        Parameters
        ----------
        quantified_variable : fe.Symbol
            universally-quantified variable

        body: fe.Expression
            first order logic proposition which to be
            satisfied by the universal quatifier.

        Returns
        -------
        fe.All
            representation of the universal predicate.
        """
        universal_predicate = (
            fe.All(
                self,
                UniversalPredicate(
                    quantified_variable.expression,
                    body.expression
                ),
                quantified_variable,
                body
            )
        )

        return universal_predicate

    @property
    def types(self) -> List[Type]:
        """
        Returns a list of the types of the symbols currently
        in the table (type can be Unknown)

        Returns
        -------
        List[Union[fe.Expression, Type]]
            List of types or Unknown if symbol has no known type
        """
        return self.symbol_table.types

    @property
    def symbol_table(self) -> TypedSymbolTable:
        """Projector to the program_ir's symbol_table"""
        return self.program_ir.symbol_table

    @property
    def symbols(self) -> Iterator[str]:
        """Iterator through the symbol's names"""
        return self._symbols_proxy

    @property
    @contextmanager
    def environment(self) -> "QuerySymbolsProxy":
        """
        Dynamic context that can be used to create
        symbols to write a Datalog program.
        Contrary to a scope, symbols stay in the symbol_table
        when exiting the environment context

        Yields
        -------
        QuerySymbolsProxy
            in dynamic mode, can be used to create symbols on-the-fly

        Example
        -------
        >>> p_ir = DatalogProgram()
        >>> nl = QueryBuilderBase(program_ir=p_ir)
        >>> with nl.environment as e:
        ...     e.x = 3
        >>> "x" in nl
        True
        >>> nl.symbols.x == 3
        True
        """
        old_dynamic_mode = self._symbols_proxy._dynamic_mode
        self._symbols_proxy._dynamic_mode = True
        try:
            yield self._symbols_proxy
        finally:
            self._symbols_proxy._dynamic_mode = old_dynamic_mode

    @property
    @contextmanager
    def scope(self) -> "QuerySymbolsProxy":
        """
        Dynamic context that can be used to create
        symbols to write a Datalog program.
        Contrary to an environment, symbols disappear from the symbol_table
        when exiting the scope context

        Yields
        -------
        QuerySymbolsProxy
            in dynamic mode, can be used to create symbols on-the-fly

        Example
        -------
        >>> p_ir = DatalogProgram()
        >>> nl = QueryBuilderBase(program_ir=p_ir)
        >>> with nl.scope as e:
        ...     e.x = 3
        >>> "x" in nl
        False
        """
        old_dynamic_mode = self._symbols_proxy._dynamic_mode
        self._symbols_proxy._dynamic_mode = True
        self.program_ir.push_scope()
        try:
            yield self._symbols_proxy
        finally:
            self.program_ir.pop_scope()
            self._symbols_proxy._dynamic_mode = old_dynamic_mode

    def new_symbol(
        self,
        type_: Union[Any, Tuple[Any, ...], List[Any]] = Unknown,
        name: str = None,
    ) -> fe.Expression:
        """
        Creates a symbol and associated expression, optionally
        specifying it's type and/or name

        Parameters
        ----------
        type_ : Union[Any, Tuple[Any, ...], List[Any]], optional
            type of the created symbol, by default Unknown
            if Iterable, will be cast to a Tuple
        name : str, optional
            name of the created symbol, by default None

        Returns
        -------
        fe.Expression
            associated to the created symbol
        """
        if isinstance(type_, (tuple, list)):
            type_ = tuple(type_)
            type_ = Tuple[type_]

        if name is None:
            name = str(uuid1())
        return fe.Expression(self, ir.Symbol[type_](name))

    @property
    def functions(self) -> List[str]:
        """
        Returns the list of symbols corresponding to callables

        Returns
        -------
        List[str]
            list of symbols of type leq Callable

        Example
        -------
        >>> p_ir = DatalogProgram()
        >>> nl = QueryBuilderBase(program_ir=p_ir)
        >>> def f(x: int) -> int:
        ...     return x+2
        >>> nl.add_symbol(f, "f")
        f: typing.Callable[[int], int] = <function f at ...>
        >>> "f" in nl.functions
        True
        >>> nl.add_symbol(3, "x")
        x: <class 'int'> = 3
        >>> "x" in nl.functions
        False
        """
        return [
            s.name
            for s in self.symbol_table
            if is_leq_informative(s.type, Callable)
        ]

    def add_symbol(
        self, value: Union[fe.Expression, ir.Constant, Any], name: str = None
    ) -> fe.Symbol:
        """
        Creates a symbol with given value and adds it to the
        current symbol_table.
        Can typicaly be used to decorate callables, or add an
        ir.Constant to the program.

        Parameters
        ----------
        value : Union[fe.Expression, ir.Constant, Any]
            value of the symbol to add. If not an fe.Expression,
            will be cast as an ir.Constant
        name : str, optional
            overrides automatic naming of the symbol, by default None

        Returns
        -------
        fe.Symbol
            created symbol

        Example
        -------
        >>> p_ir = DatalogProgram()
        >>> nl = QueryBuilderBase(program_ir=p_ir)
        >>> @nl.add_symbol
        ... def g(x: int) -> int:
        ...     return x + 2
        >>> nl.get_symbol("g")
        g: typing.Callable[[int], int] = <function g at ...>
        >>> nl.add_symbol(3, "x")
        >>> nl.get_symbol("x")
        x: <class 'int'> = 3
        """
        name = self._obtain_symbol_name(name, value)

        if isinstance(value, fe.Expression):
            value = value.expression
        elif isinstance(value, ir.Constant):
            pass
        else:
            value = ir.Constant(value)

        symbol = ir.Symbol[value.type](name)
        self.symbol_table[symbol] = value

        return fe.Symbol(self, name)

    def _obtain_symbol_name(self, name: Optional[str], value: Any) -> str:
        if name is not None:
            return name

        if not hasattr(value, "__qualname__"):
            return str(uuid1())

        if "." in value.__qualname__:
            ix = value.__qualname__.rindex(".")
            name = value.__qualname__[ix + 1 :]
        else:
            name = value.__qualname__

        return name

    def del_symbol(self, name: str) -> None:
        """
        Deletes the symbol with parameter name
        from the symbol_table

        Parameters
        ----------
        name : str
            Name of the symbol to delete

        Raises
        ------
        ValueError
            if no symbol could be found with given name

        Example
        -------
        >>> p_ir = DatalogProgram()
        >>> nl = QueryBuilderBase(program_ir=p_ir)
        >>> nl.add_symbol(3, "x")
        x: <class 'int'> = 3
        >>> nl.get_symbol("x")
        x: <class 'int'> = 3
        >>> nl.del_symbol("x")
        >>> nl.get_symbol("x")
        ValueError: Symbol x not defined
        """
        del self.symbol_table[name]

    def add_tuple_set(
        self,
        iterable: Union[Iterable, Iterable[Tuple[Any, ...]]],
        type_: Type = Unknown,
        name: str = None,
    ) -> fe.Symbol:
        """
        Creates an AbstractSet fe.Symbol containing the elements specified in
        the iterable with a List[Tuple[Any, ...]] format (see examples).
        Typically used to create extensional facts from existing databases

        Parameters
        ----------
        iterable : Union[Iterable, Iterable[Tuple[Any, ...]]]
            typically a list of tuples of values, other formats will
            be interpreted as the latter (see examples)
        type_ : Type, optional
            type of elements for the tuples, if not specified
            will be inferred from the first element, by default Unknown
        name : str, optional
            name for the AbstractSet symbol, by default None

        Returns
        -------
        fe.Symbol
            see description

        Examples
        --------
        >>> p_ir = DatalogProgram()
        >>> nl = QueryBuilderBase(program_ir=p_ir)
        >>> nl.add_tuple_set([(1, 2), (3, 4)], name="l1")
        l1: typing.AbstractSet[typing.Tuple[int, int]] = \
            [(1, 2), (3, 4)]
        >>> nl.add_tuple_set([[1, 2, 3], (3, 4)], name="l2")
        l2: typing.AbstractSet[typing.Tuple[int, int, float]] = \
            [(1, 2, 3.0), (3, 4, nan)]
        >>> nl.add_tuple_set((1, 2, 3), name="l3")
        l3: typing.AbstractSet[typing.Tuple[int]] = \
            [(1,), (2,), (3,)]
        """
        if not isinstance(type_, tuple) or len(type_) == 1:
            if isinstance(type_, tuple) and len(type_) == 1:
                type_ = type_[0]
                iterable = (e[0] for e in iterable)

            set_type = AbstractSet[type_]
        else:
            type_ = tuple(type_)
            set_type = AbstractSet[Tuple[type_]]

        constant = self._add_tuple_set_elements(iterable, set_type)
        if name is None:
            name = str(uuid1())

        symbol = ir.Symbol[set_type](name)
        self.symbol_table[symbol] = constant

        return fe.Symbol(self, name)

    def _add_tuple_set_elements(self, iterable, set_type):
        element_type = set_type.__args__[0]
        new_set = []
        for e in iterable:
            if not (isinstance(e, fe.Symbol)):
                s, c = self._create_symbol_and_get_constant(e, element_type)
                self.symbol_table[s] = c
            else:
                s = e.neurolang_symbol
            new_set.append(s)

        return ir.Constant[set_type](self.program_ir.new_set(new_set))

    @staticmethod
    def _create_symbol_and_get_constant(element, element_type):
        symbol = ir.Symbol[element_type].fresh()
        if isinstance(element, ir.Constant):
            constant = element.cast(element_type)
        elif is_leq_informative(element_type, Tuple):
            constant = ir.Constant[element_type](
                tuple(ir.Constant(ee) for ee in element)
            )
        else:
            constant = ir.Constant[element_type](element)
        return symbol, constant


class RegionMixin:
    """
    Mixin complementing a QueryBuilderBase
    with methods specific to the manipulation
    of brain volumes: regions, atlases, etc...
    """

    @property
    def region_names(self) -> List[str]:
        """
        Returns the list of symbol names with Region type

        Returns
        -------
        List[str]
            list of symbol names from symbol_table
        """
        return [s.name for s in self.symbol_table.symbols_by_type(Region)]

    @property
    def region_set_names(self) -> List[str]:
        """
        Returns the list of symbol names with set_type

        Returns
        -------
        List[str]
            list of symbol names from symbol_table
        """
        return [
            s.name for s in self.symbol_table.symbols_by_type(self.set_type)
        ]

    def new_region_symbol(self, name: Optional[str] = None) -> fe.Symbol:
        """
        Returns symbol with type Region

        Parameters
        ----------
        name : Optional[str], optional
            symbol's name, if None will be fresh, by default None

        Returns
        -------
        fe.Symbol
            see description
        """
        return self.new_symbol(type_=Region, name=name)

    def add_region(
        self, region: fe.Expression, name: Optional[str] = None
    ) -> fe.Symbol:
        """
        Adds region fe.Symbol to symbol_table

        Parameters
        ----------
        region : fe.Expression
            should be of the program_ir's type
        name : Optional[str], optional
            symbol's name, if None will be fresh, by default None

        Returns
        -------
        fe.Symbol
            symbol added to the symbol_table

        Raises
        ------
        ValueError
            if region is not of program_ir's type
        """
        if not isinstance(region, self.program_ir.type):
            raise ValueError(
                f"type mismatch between region and program_ir type:"
                f" {self.program_ir.type}"
            )

        return self.add_symbol(region, name)

    def add_region_set(
        self, region_set: Iterable, name: Optional[str] = None
    ) -> fe.Symbol:
        """
        Creates an AbstractSet fe.Symbol containing the elements specified in
        the iterable with a List[Tuple[Region]] format

        Parameters
        ----------
        region_set : Iterable
            Typically List[Tuple[Region]]
        name : Optional[str], optional
            symbol's name, if None will be fresh, by default None

        Returns
        -------
        fe.Symbol
            see description
        """
        return self.add_tuple_set(region_set, name=name, type_=Region)

    @staticmethod
    def create_region(
        spatial_image: DataobjImage,
        label: int = 1,
        prebuild_tree: bool = False,
    ) -> ExplicitVBR:
        """
        Creates an ExplicitVBR out of the voxels of a dense spatial_image
        with specified label

        Parameters
        ----------
        spatial_image : DataobjImage
            contains the .dataobj array of interest
        label : int, optional
            selects which voxel to put in the output, by default 1
        prebuild_tree : bool, optional
            see ExplicitVBR, by default False

        Returns
        -------
        ExplicitVBR
            see description
        """
        voxels = np.transpose(
            (np.asanyarray(spatial_image.dataobj) == label).nonzero()
        )
        if len(voxels) == 0:
            return None
        region = ExplicitVBR(
            voxels,
            spatial_image.affine,
            image_dim=spatial_image.shape,
            prebuild_tree=prebuild_tree,
        )
        return region

    def add_atlas_set(
        self,
        name: str,
        atlas_labels: Dict[int, str],
        spatial_image: DataobjImage,
    ) -> fe.Symbol:
        """
        Creates an atlas set:
        1- for each region specified by a label and name in atlas_labels,
        creates associated ExplicitVBR and symbols
        Tuple[region_name: str, Region]
        2- groups regions in an AbstractSet[Tuple[str, Region]] symbol with
        specified name

        Parameters
        ----------
        name : str
            name for the output atlas symbol
        atlas_labels : Dict[int, str]
            specifies which voxels to select in the spatial_image,
            and the associated region name
        spatial_image : DataobjImage
            contains the .dataobj array ofinterest

        Returns
        -------
        fe.Symbol
            see description
        """
        atlas_set = list()
        for label_number, label_name in atlas_labels.items():
            region = self.create_region(spatial_image, label=label_number)
            if region is None:
                continue
            atlas_set.append((label_name, region))
        return self.add_tuple_set(
                atlas_set,
                type_=Tuple[str, ExplicitVBR], name=name
        )

    def sphere(
        self,
        center: Union[np.ndarray, Iterable[int]],
        radius: int,
        name: Optional[str] = None,
    ) -> fe.Symbol:
        """
        Creates a Region symbol associated with the spherical
        volume described by its center and volume

        Parameters
        ----------
        center : Union[np.ndarray, Iterable[int]]
            3D center of the sphere
        radius : int
            radius of the sphere
        name : Optional[str], optional
            name of the output symbol, if None will be fresh,
            by default None

        Returns
        -------
        fe.Symbol
            see description
        """
        sr = SphericalVolume(center, radius)
        symbol = self.add_region(sr, name)
        return symbol

    def make_implicit_regions_explicit(self, affine, dim):
        """Raises NotImplementedError for now"""
        for region_symbol_name in self.region_names:
            region_symbol = self.get_symbol(region_symbol_name)
            region = region_symbol.value
            if isinstance(region, ImplicitVBR):
                self.add_region(
                    region.to_explicit_vbr(affine, dim), region_symbol_name
                )


class NeuroSynthMixin:
    """
    Neurosynth is a platform for large-scale, automated synthesis
    of functional magnetic resonance imaging (fMRI) data.
    see https://neurosynth.org/
    This Mixin complements a QueryBuilderBase with methods
    related to coordinate-based meta-analysis (CBMA) data loading.
    """

    def load_neurosynth_term_study_ids(
        self,
        term: str,
        name: Optional[str] = None,
        frequency_threshold: Optional[float] = 0.05,
    ) -> fe.Symbol:
        """
        Load study ids (PMIDs) of studies in the Neurosynth database that are
        related to a given term based on a hard thresholding of TF-IDF
        features.

        Parameters
        ----------
        term : str
            Term that studies must match.

        name : str (optional)
            Name of the symbol associated with the resulting set of studies.
            Randomly generated by default

        frequency_threshold : optional (default 0.05)
            Minimum value of the TF-IDF feature for studies to be considered
            related to the given term.

        Returns
        -------
        fe.Symbol
            fe.Symbol associated to the resulting set.

        """
        if not hasattr(self, "neurosynth_db"):
            self.neurosynth_db = NeuroSynthHandler()
        if not name:
            name = str(uuid1())
        study_set = self.neurosynth_db.ns_study_id_set_from_term(
            term, frequency_threshold
        )
        return self.add_tuple_set(
            study_set.values, type_=Tuple[StudyID], name=name
        )

    def load_neurosynth_study_tfidf_feature_for_terms(
        self, terms: Iterable[str], name: Optional[str] = None
    ) -> fe.Symbol:
        """
        Load TF-IDF features measured in each study within the Neurosynth
        database for the given terms.

        Parameters
        ----------
        terms : iterable of str
            Terms for which the TF-IDF features should be obtained.

        name : str (optional)
            Name of the symbol associated with the resulting set of (study id,
            tfidf) tuples. Randomly generated by default

        Returns
        -------
        fe.Symbol
            fe.Symbol associated to the resulting set.

        """
        if not hasattr(self, "neurosynth_db"):
            self.neurosynth_db = NeuroSynthHandler()
        if not name:
            name = str(uuid1())
        result_set = self.neurosynth_db.ns_study_tfidf_feature_for_terms(terms)
        return self.add_tuple_set(
            result_set.values, type_=Tuple[StudyID, str, TfIDf], name=name
        )

    def load_neurosynth_study_ids(
        self, name: Optional[str] = None
    ) -> fe.Symbol:
        """
        Load all study ids (PMIDs) that are part of the Neurosynth database.

        Parameters
        ----------
        name : str (optional)
            Name of the symbol associated with the resulting set of study ids.
            Randomly generated by default

        Returns
        -------
        fe.Symbol
            fe.Symbol associated to the resulting set.

        """
        if not hasattr(self, "neurosynth_db"):
            self.neurosynth_db = NeuroSynthHandler()
        if not name:
            name = str(uuid1())
        result_set = self.neurosynth_db.ns_study_ids()
        return self.add_tuple_set(result_set, type_=Tuple[StudyID], name=name)

    def load_neurosynth_reported_activations(
        self, name: Optional[str] = None
    ) -> fe.Symbol:
        """
        Load all activations reported in the Neurosynth database in the form of
        a set of (study id, voxel id) pairs.

        Parameters
        ----------
        name : str (optional)
            Name of the symbol associated with the resulting set of (study id,
            voxel id) tuples. Randomly generated by default

        Returns
        -------
        fe.Symbol
            fe.Symbol associated to the resulting set.

        """
        if not hasattr(self, "neurosynth_db"):
            self.neurosynth_db = NeuroSynthHandler()
        if not name:
            name = str(uuid1())
        result_set = self.neurosynth_db.ns_reported_activations()
        return self.add_tuple_set(
            result_set, type_=Tuple[StudyID, int], name=name
        )

    def load_neurosynth_term_study_associations(
        self,
        name: Optional[str] = None,
        threshold: Optional[float] = 1e-3,
        study_ids: Optional[List[int]] = None,
    ) -> fe.Symbol:
        """
        Load all term-to-study associations in the Neurosynth database based on
        a hard thresholding of the TF-IDF features measured from the abstracts
        of the studies.

        Parameters
        ----------
        name : str (optional)
            Name of the symbol associated with the resulting set of study ids.
            Randomly generated by default

        threshold : float (optional, default 1e-3)
            Threshold used to determine the association of a study to a given
            term.

        study_ids : iterable of int (optional)
            List of the IDs of the studies to consider. By default, all studies
            in the database are considered.

        Returns
        -------
        fe.Symbol
            fe.Symbol associated to the resulting set.

        """
        if not hasattr(self, "neurosynth_db"):
            self.neurosynth_db = NeuroSynthHandler()
        if not name:
            name = str(uuid1())
        result_set = self.neurosynth_db.ns_term_study_associations(
            threshold=threshold, study_ids=study_ids
        )
        return self.add_tuple_set(
            result_set, type_=Tuple[StudyID, str], name=name
        )


class QuerySymbolsProxy:
    """
    Class useful to create symbols on-the-fly
    Typically used in QueryBuilderBase contexts as the yielded value
    to write a program.
    Various methods are projectors to QueryBuilderBase methods
    """

    def __init__(self, query_builder):
        self._dynamic_mode = False
        self._query_builder = query_builder

    def __getattr__(self, name):
        """See QueryBuilderBase.get_symbol"""
        if name in self.__getattribute__("_query_builder"):
            return self._query_builder.get_symbol(name)

        try:
            return super().__getattribute__(name)
        except AttributeError:
            if self._dynamic_mode:
                return self._query_builder.new_symbol(Unknown, name=name)
            else:
                raise

    def __setattr__(self, name, value):
        """See QueryBuilderBase.add_symbol"""
        if name == "_dynamic_mode":
            return super().__setattr__(name, value)
        elif self._dynamic_mode:
            return self._query_builder.add_symbol(value, name=name)
        else:
            return super().__setattr__(name, value)

    def __delattr__(self, name):
        """See QueryBuilderBase.del_symbol"""
        if self._dynamic_mode and name:
            self._query_builder.del_symbol(name)
        else:
            super().__delattr__(name)

    def __getitem__(self, attr):
        """See QueryBuilderBase.get_symbol"""
        return self._query_builder.get_symbol(attr)

    def __setitem__(self, key, value):
        """See QueryBuilderBase.add_symbol"""
        return self._query_builder.add_symbol(value, name=key)

    def __contains__(self, symbol):
        """See QueryBuilderBase.__contains__"""
        return symbol in self._query_builder.symbol_table

    def __iter__(self) -> List[str]:
        """
        Iterates through the names of the symbols
        currently in the symbol_table, ordered in ascending name

        Returns
        -------
        List[str]
            list of symbol names
        """
        return iter(
            sorted(set(s.name for s in self._query_builder.symbol_table))
        )

    def __len__(self) -> int:
        """
        Returns number of symbols currently in symbol_table

        Returns
        -------
        int
            see description
        """
        return len(self._query_builder.symbol_table)

    def __dir__(self):
        """Descibes self and lists symbols in current symbol_table"""
        init = object.__dir__(self)
        init += [symbol.name for symbol in self._query_builder.symbol_table]
        return init

    def __repr__(self):
        """Describes symbols currently in symbol_table"""
        init = [symbol.name for symbol in self._query_builder.symbol_table]
        return f"QuerySymbolsProxy with symbols {init}"



from ...expressions import (
    Expression,
    Symbol,
    FunctionApplication as Fa,
    Constant as C,
)
from ...expression_walker import (
    add_match,
    ExpressionWalker,
    ReplaceSymbolWalker,
)
from .chart_parser import Quote, CODE_QUOTE
from .english_grammar import S, V, NP, VP, PN, DET, N, VAR, SL, LIT
from .exceptions import ParseDatalogPredicateException
from ...logic import (
    Implication,
    Conjunction,
    ExistentialPredicate,
    UniversalPredicate,
)
import re


def indent(s, tab="    "):
    return "".join(tab + l for l in s.splitlines(keepend=True))


class DRS(Expression):
    def __init__(self, referents, expressions):
        self.referents = referents
        self.expressions = expressions

    def __repr__(self):
        return (
            "DRS <"
            + ", ".join(map(repr, self.referents))
            + "> [\n"
            + "".join("    " + repr(e) + ",\n" for e in self.expressions)
            + "]"
        )


class DRSBuilder(ExpressionWalker):
    def __init__(self, grammar):
        self.grammar = grammar

    @add_match(
        DRS, lambda drs: any(isinstance(e, DRS) for e in drs.expressions),
    )
    def join_drs(self, drs):
        refs = drs.referents
        exps = ()
        for e in drs.expressions:
            if isinstance(e, DRS):
                refs += tuple(r for r in e.referents if r not in refs)
                exps += e.expressions
            else:
                exps += (e,)
        return self.walk(DRS(refs, exps))

    @add_match(Fa, lambda fa: any(isinstance(e, DRS) for e in fa.args))
    def float_drs(self, fa):
        args = ()
        drs = None
        for e in fa.args:
            if isinstance(e, DRS) and not drs:
                drs = e
                args += (drs.expressions[0],)
            else:
                args += (e,)
        exps = (Fa(fa.functor, args),) + tuple(drs.expressions[1:])
        return self.walk(DRS(drs.referents, exps))

    @add_match(Fa(Fa(NP, ...), (Fa(Fa(PN, ...), ...),)))
    def proper_names(self, np):
        (pn,) = np.args
        (_, _, const) = pn.functor.args
        return self.walk(DRS((), (const,)))

    @add_match(
        Fa(Fa(S, ...), (..., Fa(Fa(VP, ...), (Fa(Fa(V, ...), ...), ...)),),)
    )
    def predicate(self, s):
        (subject, vp) = s.args
        (v, object_) = vp.args
        exp = Symbol(v.args[0].value)(subject, object_)
        return self.walk(DRS((), (exp,)))

    @add_match(
        Fa(Fa(NP, ...), (Fa(Fa(DET, ...), ...), Fa(Fa(N, ...), ...),)),
        lambda np: np.args[0].args[0].value in ["a", "an"],
    )
    def indefinite_noun_phrase(self, np):
        (det, n) = np.args
        x = Symbol.fresh()
        exp = Symbol(n.args[0].value)(x)
        return self.walk(DRS((x,), (x, exp)))

    @add_match(Fa(Fa(NP, ...), (Fa(Fa(VAR, ...), ...),)),)
    def var_noun_phrase(self, np):
        (var,) = np.args
        v = Symbol(var.args[0].value)
        return self.walk(DRS((v,), (v,)))

    @add_match(
        Fa(Fa(NP, ...), (Fa(Fa(NP, ...), ...), Fa(Fa(VAR, ...), ...),)),
    )
    def var_apposition(self, np):
        (np, var) = np.args
        np_drs = self.walk(np)

        y = Symbol(var.args[0].value)
        x = np_drs.expressions[0]
        rsw = ReplaceSymbolWalker({x: y})

        exps = ()
        for e in np_drs.expressions:
            exps += (rsw.walk(e),)

        refs = ()
        for r in np_drs.referents:
            refs += (rsw.walk(r),)

        return self.walk(DRS(refs, exps))

    @add_match(
        Fa(
            Fa(S, ...),
            (C("if"), Fa(Fa(S, ...), ...), C("then"), Fa(Fa(S, ...), ...),),
        ),
    )
    def conditional(self, s):
        (_, ant, _, cons) = s.args
        return self.walk(DRS((), (Implication(cons, ant),)))

    @add_match(Fa(Fa(S, ...), (Fa(Quote, (C(CODE_QUOTE), ...)),),),)
    def quoted_predicate(self, s):
        exp = _parse_predicate(s.args[0].args[1].value)
        return self.walk(DRS(exp.args, (exp,)))

    @add_match(
        Implication(DRS, DRS),
        lambda impl: (
            set(impl.antecedent.referents) & set(impl.consequent.referents)
        ),
    )
    def implication(self, impl):
        drs_ant = impl.antecedent
        drs_con = impl.consequent
        drs_con.referents = tuple(
            set(drs_con.referents) - set(drs_ant.referents)
        )
        return self.walk(Implication(drs_con, drs_ant))

    @add_match(
        Fa(Fa(S, ...), (Fa(Fa(S, ...), ...), C("and"), Fa(Fa(S, ...), ...),),),
    )
    def simple_and(self, s):
        (a, _, b) = s.args
        a = self.walk(a)
        b = self.walk(b)
        return self.walk(DRS((), (a, b,)))

    @add_match(
        Fa(
            Fa(S, ...),
            (Fa(Fa(SL, ...), ...), C(","), C("and"), Fa(Fa(S, ...), ...),),
        ),
    )
    def comma_and(self, s):
        (sl, _, _, s) = s.args
        sl = self.walk(sl)
        s = self.walk(s)
        return self.walk(DRS((), sl + (s,)))

    @add_match(
        Fa(
            Fa(SL, ...),
            (Fa(Fa(S, ...), ...),),
        ),
    )
    def single_sentence_list(self, sl):
        (s,) = sl.args
        return (self.walk(s),)

    @add_match(
        Fa(
            Fa(SL, ...),
            (Fa(Fa(SL, ...), ...), C(","), Fa(Fa(S, ...), ...)),
        ),
    )
    def sentence_list(self, sl):
        (sl, _, s) = sl.args
        sl = self.walk(sl)
        s = self.walk(s)
        return sl + (s,)

    @add_match(Fa(Fa(NP, ...), (Fa(Fa(LIT, ...), ...),)),)
    def lit_noun_phrase(self, np):
        (lit,) = np.args
        (const,) = lit.functor.args
        return self.walk(DRS((), (const,)))


r = re.compile(r"^(\w+)\((\w+(,\s\w+)*)\)$")


def _parse_predicate(string):
    # This could totally use the datalog parser
    m = r.match(string)
    if not m:
        raise ParseDatalogPredicateException(
            f"Quoted predicate is not valid datalog: {string}"
        )
    functor = Symbol(m.group(1))
    args = map(Symbol, map(str.strip, m.group(2).split(",")))
    return functor(*args)


class DRS2FOL(ExpressionWalker):
    @add_match(DRS)
    def drs(self, drs):
        exp = Conjunction(tuple(map(self.walk, drs.expressions)))
        for r in drs.referents:
            exp = ExistentialPredicate(r, exp)
        return self.walk(exp)

    @add_match(Conjunction((...,)))
    def unary_conjunction(self, conj):
        return self.walk(conj.formulas[0])

    @add_match(Implication(DRS, DRS))
    def implication(self, impl):
        drs_ant = impl.antecedent
        drs_con = impl.consequent
        ant = Conjunction(tuple(map(self.walk, drs_ant.expressions)))
        con = self.walk(drs_con)
        exp = Implication(con, ant)
        for r in drs_ant.referents:
            exp = UniversalPredicate(r, exp)
        return self.walk(exp)



from operator import contains, eq, gt, mul, not_
from typing import AbstractSet, Tuple

import pytest

from ...exceptions import NeuroLangException
from ...expressions import Constant, FunctionApplication, Symbol
from ...logic import Disjunction
from ...relational_algebra import (
    ColumnInt,
    ColumnStr,
    Destroy,
    Difference,
    ExtendedProjection,
    ExtendedProjectionListMember,
    NameColumns,
    NaturalJoin,
    Projection,
    Selection,
    Union,
    str2columnstr_constant,
)
from ...utils import NamedRelationalAlgebraFrozenSet
from ..expressions import Conjunction, Negation
from ..translate_to_named_ra import TranslateToNamedRA

C_ = Constant
S_ = Symbol
F_ = FunctionApplication

EQ = Constant(eq)


def test_translate_set():
    x = S_("x")
    y = S_("y")
    R1 = S_("R1")
    fa = R1(x, y)

    tr = TranslateToNamedRA()
    res = tr.walk(fa)
    assert res == NameColumns(
        Projection(R1, (C_(ColumnInt(0)), C_(ColumnInt(1)))),
        (Constant(ColumnStr("x")), Constant(ColumnStr("y"))),
    )

    fa = R1(C_(1), y)

    tr = TranslateToNamedRA()
    res = tr.walk(fa)
    assert res == NameColumns(
        Projection(
            Selection(R1, C_(eq)(C_(ColumnInt(0)), C_(1))), (C_(ColumnInt(1)),)
        ),
        (Constant(ColumnStr("y")),),
    )


def test_equality_constant_symbol():
    x = S_("x")
    a = C_("a")
    R1 = S_("R1")

    expected_result = C_[AbstractSet[Tuple[str]]](
        NamedRelationalAlgebraFrozenSet(("x",), {"a"})
    )

    fa = C_(eq)(x, a)
    tr = TranslateToNamedRA()
    res = tr.walk(Conjunction((fa,)))
    assert res == expected_result

    fa = C_(eq)(a, x)
    tr = TranslateToNamedRA()
    res = tr.walk(Conjunction((fa,)))
    assert res == expected_result

    y = S_("y")
    fb = R1(x, y)

    exp = Conjunction((fb, fa))

    expected_result = Selection(
        NameColumns(
            Projection(R1, (C_(ColumnInt(0)), C_(ColumnInt(1)))),
            (Constant(ColumnStr("x")), Constant(ColumnStr("y"))),
        ),
        C_(eq)(C_(ColumnStr("x")), a),
    )

    res = tr.walk(exp)
    assert res == expected_result


def test_equality_symbols():
    x = S_("x")
    y = S_("y")
    z = S_("z")
    w = S_("w")
    R1 = S_("R1")

    y = S_("y")
    fb = R1(x, y)
    fb_trans = NameColumns(
        Projection(R1, (C_(ColumnInt(0)), C_(ColumnInt(1)))),
        (Constant(ColumnStr("x")), Constant(ColumnStr("y"))),
    )

    exp = Conjunction((fb, C_(eq)(x, y)))

    expected_result = Selection(
        fb_trans, C_(eq)(C_(ColumnStr("x")), C_(ColumnStr("y")))
    )

    tr = TranslateToNamedRA()
    res = tr.walk(exp)
    assert res == expected_result

    exp = Conjunction((fb, C_(eq)(x, z)))

    expected_result = ExtendedProjection(
        fb_trans,
        (
            ExtendedProjectionListMember(
                Constant(ColumnStr("x")), Constant(ColumnStr("x"))
            ),
            ExtendedProjectionListMember(
                Constant(ColumnStr("y")), Constant(ColumnStr("y"))
            ),
            ExtendedProjectionListMember(
                Constant(ColumnStr("x")), Constant(ColumnStr("z"))
            ),
        ),
    )

    res = tr.walk(exp)
    assert res == expected_result

    exp = Conjunction((fb, C_(eq)(z, x)))

    expected_result = ExtendedProjection(
        fb_trans,
        (
            ExtendedProjectionListMember(
                Constant(ColumnStr("x")), Constant(ColumnStr("x"))
            ),
            ExtendedProjectionListMember(
                Constant(ColumnStr("y")), Constant(ColumnStr("y"))
            ),
            ExtendedProjectionListMember(
                Constant(ColumnStr("x")), Constant(ColumnStr("z"))
            ),
        ),
    )

    res = tr.walk(exp)
    assert res == expected_result

    exp = Conjunction((fb, C_(eq)(z, w)))
    with pytest.raises(NeuroLangException, match="Could not resolve*"):
        res = tr.walk(exp)


def test_joins():
    x = S_("x")
    y = S_("y")
    z = S_("z")
    R1 = S_("R1")
    fa = R1(x, y)
    fb = R1(y, z)
    exp = Conjunction((fa, fb))

    fa_trans = NameColumns(
        Projection(R1, (C_(ColumnInt(0)), C_(ColumnInt(1)))),
        (Constant(ColumnStr("x")), Constant(ColumnStr("y"))),
    )

    fb_trans = NameColumns(
        Projection(R1, (C_(ColumnInt(0)), C_(ColumnInt(1)))),
        (Constant(ColumnStr("y")), Constant(ColumnStr("z"))),
    )

    tr = TranslateToNamedRA()
    res = tr.walk(exp)
    assert res == NaturalJoin(fa_trans, fb_trans)

    R2 = S_("R2")
    fb = R2(x, y)
    fb_trans = NameColumns(
        Projection(R2, (C_(ColumnInt(0)), C_(ColumnInt(1)))),
        (Constant(ColumnStr("x")), Constant(ColumnStr("y"))),
    )
    exp = Conjunction((fa, Negation(fb)))

    tr = TranslateToNamedRA()
    res = tr.walk(exp)
    assert res == Difference(fa_trans, fb_trans)

    fa = R1(x, y)
    fb = R2(y, C_(0))
    fb_trans = NameColumns(
        Projection(
            Selection(R2, C_(eq)(C_(ColumnInt(1)), C_(0))), (C_(ColumnInt(0)),)
        ),
        (Constant(ColumnStr("y")),),
    )

    exp = Conjunction((fa, Negation(fb)))

    tr = TranslateToNamedRA()
    res = tr.walk(exp)

    assert res == Difference(fa_trans, NaturalJoin(fa_trans, fb_trans))


def test_selection():
    x = S_("x")
    y = S_("y")
    R1 = S_("R1")
    fa = R1(x, y)
    builtin_condition = C_(gt)(x, C_(3))
    exp = Conjunction((fa, builtin_condition))

    tr = TranslateToNamedRA()
    res = tr.walk(exp)
    fa_trans = NameColumns(
        Projection(R1, (C_(ColumnInt(0)), C_(ColumnInt(1)))),
        (Constant(ColumnStr("x")), Constant(ColumnStr("y"))),
    )
    assert res == Selection(fa_trans, builtin_condition)


def test_extended_projection():
    x = S_("x")
    y = S_("y")
    z = S_("z")
    R1 = S_("R1")
    fa = R1(x, y)
    builtin_condition = C_(eq)(C_(mul)(x, C_(3)), z)
    exp = Conjunction((fa, builtin_condition))

    tr = TranslateToNamedRA()
    res = tr.walk(exp)
    fa_trans = NameColumns(
        Projection(R1, (C_(ColumnInt(0)), C_(ColumnInt(1)))),
        (Constant(ColumnStr("x")), Constant(ColumnStr("y"))),
    )
    exp_trans = ExtendedProjection(
        fa_trans,
        [
            ExtendedProjectionListMember(*builtin_condition.args),
            ExtendedProjectionListMember(x, x),
            ExtendedProjectionListMember(y, y),
        ],
    )
    assert res == exp_trans


def test_extended_projection_2():
    u = S_("u")
    v = S_("v")
    w = S_("w")
    x = S_("x")
    y = S_("y")
    z = S_("z")
    R1 = S_("R1")
    fa = R1(x, y, v, u)
    builtin_condition_1 = C_(eq)(C_(mul)(x, C_(3)), z)
    builtin_condition_2 = C_(eq)(C_(mul)(v, C_(2)), w)
    exp = Conjunction((fa, builtin_condition_1, builtin_condition_2))

    tr = TranslateToNamedRA()
    res = tr.walk(exp)
    fa_trans = NameColumns(
        Projection(
            R1,
            (
                C_(ColumnInt(0)),
                C_(ColumnInt(1)),
                C_(ColumnInt(2)),
                C_(ColumnInt(3)),
            ),
        ),
        (
            Constant(ColumnStr("x")),
            Constant(ColumnStr("y")),
            Constant(ColumnStr("v")),
            Constant(ColumnStr("u")),
        ),
    )
    exp_trans = ExtendedProjection(
        fa_trans,
        [
            ExtendedProjectionListMember(*builtin_condition_1.args),
            ExtendedProjectionListMember(*builtin_condition_2.args),
            ExtendedProjectionListMember(x, x),
            ExtendedProjectionListMember(y, y),
            ExtendedProjectionListMember(u, u),
            ExtendedProjectionListMember(v, v),
        ],
    )
    assert res == exp_trans


def test_extended_projection_algebraic_expression():
    x = S_("x")
    y = S_("y")
    R1 = S_("R1")
    fa = R1(x, y)
    builtin_condition = C_(eq)(C_(mul)(C_(2), C_(3)), y)
    exp = Conjunction((fa, builtin_condition))

    tr = TranslateToNamedRA()
    res = tr.walk(exp)
    fa_trans = NameColumns(
        Projection(R1, (C_(ColumnInt(0)), C_(ColumnInt(1)))),
        (Constant(ColumnStr("x")), Constant(ColumnStr("y"))),
    )
    assert res == Selection(
        fa_trans, Constant(eq)(Constant(ColumnStr("y")), Constant(6))
    )


def test_set_destroy():
    r1 = S_("R1")
    x = S_("x")
    y = S_("y")

    tr = TranslateToNamedRA()

    exp = Conjunction((C_(contains)(x, y), r1(x)))
    res = tr.walk(exp)

    exp_result = Destroy(
        NameColumns(Projection(r1, (C_(0),)), (C_("x"),)), x, y
    )
    assert res == exp_result


def test_set_destroy_multicolumn():
    r1 = S_("R1")
    x = S_("x")
    y = S_("y")
    z = S_("z")

    tr = TranslateToNamedRA()
    exp = Conjunction((C_(contains)(x, C_((y, z))), r1(x)))
    res = tr.walk(exp)

    exp_result = Destroy(
        NameColumns(Projection(r1, (C_(0),)), (C_("x"),)),
        x,
        C_[Tuple[ColumnStr, ColumnStr]]((ColumnStr("y"), ColumnStr("z"))),
    )
    assert res == exp_result


def test_set_constant_contains():
    r1 = S_("R1")
    x = S_("x")
    exp = Conjunction((C_(contains)(x, C_(0)), r1(x)))

    tr = TranslateToNamedRA()
    res = tr.walk(exp)

    exp_result = Selection(
        NameColumns(Projection(r1, (C_(0),)), (C_("x"),)),
        C_(contains)(x, C_(0)),
    )
    assert res == exp_result


def test_only_equality():
    exp = Conjunction((C_(eq)(S_("x"), C_(3)),))

    res = TranslateToNamedRA().walk(exp)

    exp_result = NamedRelationalAlgebraFrozenSet(("x",), (3,))

    assert isinstance(res, Constant)
    assert res.value == exp_result


def test_disjunction():
    p1 = S_("T")(S_("x"))
    p2 = S_("U")(S_("x"))
    p3 = S_("V")(S_("x"))
    r1 = TranslateToNamedRA().walk(p1)
    r2 = TranslateToNamedRA().walk(p2)
    r3 = TranslateToNamedRA().walk(p3)

    exp = Disjunction((p1,))
    res = TranslateToNamedRA().walk(exp)
    assert res == r1

    exp = Disjunction((p1, p2))
    res = TranslateToNamedRA().walk(exp)
    assert res == Union(r1, r2)

    exp = Disjunction((p1, p2, p3))
    res = TranslateToNamedRA().walk(exp)
    assert res == Union(r1, Union(r2, r3))

    exp = Disjunction((p1, Conjunction((p2, p3))))
    res = TranslateToNamedRA().walk(exp)
    assert res == Union(r1, TranslateToNamedRA().walk(Conjunction((p2, p3))))

    exp = Conjunction((p1, Disjunction((p2, p3))))
    res = TranslateToNamedRA().walk(exp)
    assert res == NaturalJoin(
        r1, TranslateToNamedRA().walk(Disjunction((p2, p3)))
    )


def test_border_cases():
    R1 = S_("R1")
    x = S_("x")
    y = S_("y")
    z = S_("z")

    tr = TranslateToNamedRA()

    exp = Conjunction((R1(x), C_(eq)(x, x)))
    res = tr.walk(exp)

    assert res == NameColumns(
        Projection(R1, (C_(0),)), (Constant(ColumnStr("x")),)
    )

    exp = Conjunction((Negation(Negation(R1(x))),))
    res = tr.walk(exp)

    assert res == NameColumns(
        Projection(R1, (C_(0),)), (Constant(ColumnStr("x")),)
    )

    tr = TranslateToNamedRA()

    exp = Conjunction((R1(x), C_(eq)(x, y), C_(eq)(z, C_(2) * y)))
    res = tr.walk(exp)
    expected_res = ExtendedProjection(
        ExtendedProjection(
            NameColumns(Projection(R1, (C_(0),)), (C_("x"),)),
            (
                ExtendedProjectionListMember(
                    C_(ColumnStr("x")),
                    C_(ColumnStr("x")),
                ),
                ExtendedProjectionListMember(
                    C_(ColumnStr("x")),
                    C_(ColumnStr("y")),
                ),
            ),
        ),
        (
            ExtendedProjectionListMember(C_("x"), C_("x")),
            ExtendedProjectionListMember(C_("y"), C_("y")),
            ExtendedProjectionListMember(C_(2) * C_("y"), C_("z")),
        ),
    )
    assert res == expected_res


def test_border_case_2():
    T = Symbol[AbstractSet[int]]("T")
    x = Symbol[int]("x")
    y = Symbol[int]("y")

    def gtz_f(x):
        return x > 0

    gtz = Constant(gtz_f)

    exp = Conjunction((T(x), Negation(gtz(x))))

    expected_res = Selection(
        NameColumns(Projection(T, (C_(0),)), (C_(ColumnStr("x")),)),
        C_(not_)(gtz(C_(ColumnStr("x")))),
    )

    res = TranslateToNamedRA().walk(exp)
    assert res == expected_res

    exp = Conjunction((T(x), Negation(C_(eq)(x, C_(3)))))

    res = TranslateToNamedRA().walk(exp)

    expected_res = Selection(
        NameColumns(Projection(T, (C_(0),)), (C_(ColumnStr("x")),)),
        C_(not_)(C_(eq)(C_(ColumnStr("x")), C_(3))),
    )
    assert res == expected_res

    exp = Conjunction((T(x, y), Negation(C_(eq)(x, y))))

    res = TranslateToNamedRA().walk(exp)

    expected_res = Selection(
        NameColumns(Projection(T, (C_(0), C_(1))), (C_("x"), C_("y"))),
        C_(not_)(C_(eq)(C_(ColumnStr("x")), C_(ColumnStr("y")))),
    )
    assert res == expected_res


def test_extended_projection_variable_equality():
    Q = Symbol("Q")
    x = Symbol("x")
    y = Symbol("y")
    conjunction = Conjunction((Q(x), EQ(y, x)))
    result = TranslateToNamedRA().walk(conjunction)
    assert result == ExtendedProjection(
        NameColumns(
            Projection(Q, (C_(ColumnInt(0)),)),
            (C_(ColumnStr("x")),),
        ),
        (
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("x"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("y"),
            ),
        ),
    )


def test_extended_projection_variable_equality_constant():
    Q = Symbol("Q")
    x = Symbol("x")
    y = Symbol("y")
    a = Constant("a")
    conjunction = Conjunction((Q(x), EQ(y, a)))
    result = TranslateToNamedRA().walk(conjunction)
    assert result == ExtendedProjection(
        NameColumns(
            Projection(Q, (C_(ColumnInt(0)),)),
            (C_(ColumnStr("x")),),
        ),
        (
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("x"),
            ),
            ExtendedProjectionListMember(
                Constant("a"),
                str2columnstr_constant("y"),
            ),
        ),
    )


def test_extended_projection_variable_equality_not_named():
    Q = Symbol("Q")
    x = Symbol("x")
    y = Symbol("y")
    z = Symbol("z")
    conjunction = Conjunction((Q(x), EQ(y, x), EQ(z, y)))
    result = TranslateToNamedRA().walk(conjunction)
    expected = ExtendedProjection(
        NameColumns(
            Projection(Q, (C_(ColumnInt(0)),)),
            (C_(ColumnStr("x")),),
        ),
        (
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("x"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("y"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("y"),
                str2columnstr_constant("z"),
            ),
        ),
    )
    assert result == expected


def test_extended_projection_variable_equality_twice_same_lhs():
    Q = Symbol("Q")
    x = Symbol("x")
    y = Symbol("y")
    z = Symbol("z")
    conjunction = Conjunction((Q(x), EQ(x, y), EQ(x, z)))
    result = TranslateToNamedRA().walk(conjunction)
    assert result == ExtendedProjection(
        NameColumns(
            Projection(Q, (C_(ColumnInt(0)),)),
            (C_(ColumnStr("x")),),
        ),
        (
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("x"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("y"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("z"),
            ),
        ),
    )


def test_double_equality():
    Q = Symbol("Q")
    x = Symbol("x")
    y = Symbol("y")
    z = Symbol("z")
    conjunction = Conjunction((Q(x), EQ(x, y), EQ(x, z)))
    result = TranslateToNamedRA().walk(conjunction)
    expected = ExtendedProjection(
        NameColumns(
            Projection(Q, (C_(ColumnInt(0)),)),
            (C_(ColumnStr("x")),),
        ),
        (
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("x"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("z"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("y"),
            ),
        ),
    )
    assert result == expected


def test_three_way_equality():
    Q = Symbol("Q")
    x = Symbol("x")
    y = Symbol("y")
    w = Symbol("w")
    z = Symbol("z")
    conjunction = Conjunction((Q(x), EQ(z, y), EQ(y, w), EQ(w, x)))
    result = TranslateToNamedRA().walk(conjunction)
    expected = ExtendedProjection(
        NameColumns(
            Projection(Q, (C_(ColumnInt(0)),)),
            (C_(ColumnStr("x")),),
        ),
        (
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("x"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("w"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("w"),
                str2columnstr_constant("y"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("y"),
                str2columnstr_constant("z"),
            ),
        ),
    )
    assert result == expected


def test_not_ordered_equalites():
    Q = Symbol("Q")
    x = Symbol("x")
    y = Symbol("y")
    w = Symbol("w")
    h = Symbol("h")
    z = Symbol("z")
    conjunction = Conjunction((Q(x), EQ(w, h), EQ(h, x), EQ(y, w), EQ(z, y)))
    result = TranslateToNamedRA().walk(conjunction)
    expected = ExtendedProjection(
        NameColumns(
            Projection(Q, (C_(ColumnInt(0)),)),
            (C_(ColumnStr("x")),),
        ),
        (
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("x"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("h"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("h"),
                str2columnstr_constant("w"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("w"),
                str2columnstr_constant("y"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("y"),
                str2columnstr_constant("z"),
            ),
        ),
    )
    assert result == expected


def test_equality_builtin_application():
    x = Symbol("x")
    y = Symbol("y")
    z = Symbol("z")
    P = Symbol("P")
    f = Constant(lambda x: x ** 2)
    conjunction = Conjunction((P(x), P(y), EQ(z, f(x, y))))
    result = TranslateToNamedRA().walk(conjunction)
    expected = ExtendedProjection(
        NaturalJoin(
            NameColumns(
                Projection(P, (C_(ColumnInt(0)),)),
                (C_(ColumnStr("x")),),
            ),
            NameColumns(
                Projection(P, (C_(ColumnInt(0)),)),
                (C_(ColumnStr("y")),),
            ),
        ),
        (
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("x"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("y"),
                str2columnstr_constant("y"),
            ),
            ExtendedProjectionListMember(
                f(C_(ColumnStr("x")), C_(ColumnStr("y"))),
                str2columnstr_constant("z"),
            ),
        ),
    )
    assert result == expected


def test_equality_nested_builtin_application():
    x = Symbol("x")
    y = Symbol("y")
    z = Symbol("z")
    P = Symbol("P")
    f = Constant(lambda x: x // 2)
    g = Constant(lambda x: x ** 2)
    conjunction = Conjunction((P(x), P(y), EQ(z, f(g(x), g(y)))))
    result = TranslateToNamedRA().walk(conjunction)
    expected = ExtendedProjection(
        NaturalJoin(
            NameColumns(
                Projection(P, (C_(ColumnInt(0)),)),
                (C_(ColumnStr("x")),),
            ),
            NameColumns(
                Projection(P, (C_(ColumnInt(0)),)),
                (C_(ColumnStr("y")),),
            ),
        ),
        (
            ExtendedProjectionListMember(
                str2columnstr_constant("x"),
                str2columnstr_constant("x"),
            ),
            ExtendedProjectionListMember(
                str2columnstr_constant("y"),
                str2columnstr_constant("y"),
            ),
            ExtendedProjectionListMember(
                f(g(C_(ColumnStr("x"))), g(C_(ColumnStr("y")))),
                str2columnstr_constant("z"),
            ),
        ),
    )
    assert result == expected
